# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations
import copy
import json
import logging
import multiprocessing as mp
import numbers
import os
import typing as t_
from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum

import h5py
import numpy as np
import pandas as pd
import tifffile as tf
from matplotlib import pyplot as plt, widgets
from scipy import interpolate as spi
from scipy.io import savemat
from . import _metadata as pwsdtmd
from . import _other
if t_.TYPE_CHECKING:
    from ..utility.reflection import Material


class ICBase(ABC):
    """A class to handle the data operations common to PWS related `image cubes`. Does not contain any file specific
    functionality. uses the generic `index` attribute which can be overridden by derived classes to be wavelength, wavenumber,
    time, etc.

    Args:
        data (np.ndarray): A 3-dimensional array containing the data the dimensions should be [Y, X, Z] where X and Y are the spatial coordinates of the image
            and Z corresponds to the `index` dimension, e.g. wavelength, wavenumber, time, etc.
        index (tuple(Number)): A tuple containing the values of the index for the data. This could be a tuple of wavelength values, times (in the case of Dyanmics), etc.
        dtype (type): the data type that the data should be stored as. The default is numpy.float32.
    """
    _index: tuple
    data: np.ndarray

    def __init__(self, data: np.ndarray, index: tuple, dtype=np.float32):
        assert isinstance(data, np.ndarray)
        self.data = data.astype(dtype)
        self._index = index
        if self.data.shape[2] != len(self.index):
            raise ValueError(f"The length of the index list doesn't match the index axis of the data array. Got {len(self.index)}, expected {self.data.shape[2]}.")

    @property
    @abstractmethod
    def _hdfTypeName(self) -> str:
        """Each class of this type should have a unique constant name which will be used to identify it when saved as HDF."""
        pass

    @property
    def index(self) -> t_.Tuple[float, ...]:
        """

        Returns:
            The values of the datacube's index
        """
        return self._index


    def plotMean(self) -> t_.Tuple[plt.Figure, plt.Axes]:
        """

        Returns:
            A figure and attached axes plotting the mean of the data along the index axis.
                corresponds to the mean reflectance in most cases.
        """
        fig, ax = plt.subplots()
        mean = np.mean(self.data, axis=2)
        im = ax.imshow(mean)
        plt.colorbar(im, ax=ax)
        return fig, ax

    def getMeanSpectra(self, mask: t_.Optional[t_.Union[_other.Roi, np.ndarray]] = None) -> t_.Tuple[np.ndarray, np.ndarray]:
        """Calculate the average spectra within a region of the data.

        Args:
            mask: An optional other.Roi or boolean numpy array used to select pixels from the X and Y dimensions of the data array.
                If left as None then the full data array will be used as the region.

        Returns:
            The average spectra within the region, the standard deviation of the spectra within the region
        """
        if isinstance(mask, _other.Roi):
            mask = mask.mask
        if mask is None: #Make a mask that includes everything
            mask = np.ones(self.data.shape[:-1], dtype=np.bool)
        mean = self.data[mask].mean(axis=0)
        std = self.data[mask].std(axis=0)
        return mean, std

    def selectLassoROI(self, displayIndex: t_.Optional[int] = None, clim: t_.Sequence = None) -> _other.Roi:
        """
        Allow the user to draw a `freehand` ROI on an image of the acquisition.

        Args:
            displayIndex: Display a particular z-slice of the array for mask drawing. If `None` then the mean along Z is displayed.

        Returns:
            An array of vertices of the polygon drawn.
        """
        import warnings
        warnings.warn("This method has been moved to the `pwspy_gui.utility` module and will be removed in the future.", category=DeprecationWarning)
        Verts = [None]
        if displayIndex is None:
            displayIndex = self.data.shape[2]//2
        fig, ax = plt.subplots()
        data = self.data[:, :, displayIndex]
        ax.imshow(data, clim=[np.percentile(data, 1), np.percentile(data, 99)])
        fig.suptitle("Close to accept ROI")

        def onSelect(verts):
            Verts[0] = verts

        l = widgets.LassoSelector(ax, onSelect, lineprops={'color': 'r'})
        fig.show()
        while plt.fignum_exists(fig.number):
            fig.canvas.flush_events()
        return _other.Roi.fromVerts(np.array(Verts[0]), self.data.shape[:2])

    def selectRectangleROI(self, displayIndex: t_.Optional[int] = None) -> np.ndarray:
        """
        Allow the user to draw a rectangular ROI on an image of the acquisition.

        Args:
            displayIndex (int): is used to display a particular z-slice for mask drawing. If None then the mean along Z is displayed. Returns an array of vertices of the rectangle.

        Returns:
            np.ndarray: An array of the 4 XY vertices of the rectangle.
        """
        import warnings
        warnings.warn("This method has been moved to the `pwspy_gui.utility` module and will be removed in the future.",
                      category=DeprecationWarning)

        verts = [None]

        if displayIndex is None:
           displayIndex = self.data.shape[2]//2
        fig, ax = plt.subplots()
        ax.imshow(self.data[:, :, displayIndex])
        fig.suptitle("Close to accept ROI")

        def rectSelect(mins, maxes):
            verts[0] = ((mins.ydata, mins.xdata), (maxes.ydata, mins.xdata), (maxes.ydata, maxes.xdata), (mins.ydata, maxes.xdata))

        r = widgets.RectangleSelector(ax, rectSelect)
        fig.show()
        while plt.fignum_exists(fig.number):
            fig.canvas.flush_events()
        return np.array(verts[0])

    def __getitem__(self, slic):
        return self.data[slic]

    def filterDust(self, sigma: float, pixelSize: float):
        """Blurs the data cube in the X and Y dimensions. Often used to remove the effects of dust on a normalization.

        Args:
            sigma: This specifies the radius of the gaussian filter used for blurring. The units of the value are determined by `pixelSize`
            pixelSize: The pixel size in microns. Settings this to 1 will effectively causes sigma to be in units of pixels rather than microns."""
        from scipy import ndimage
        sigma = sigma / pixelSize  # convert from microns to pixels
        for i in range(self.data.shape[2]):
            self.data[:, :, i] = ndimage.filters.gaussian_filter(self.data[:, :, i], sigma, mode='reflect')

    def _indicesMatch(self, other: 'ICBase') -> bool:
        """This check is performed before allowing many arithmetic operations between two data cubes. Makes sure that the Z-axis of the two cubes match."""
        return self._index == other._index

    def selIndex(self, start: t_.Optional[float], stop: t_.Optional[float]) -> t_.Tuple[np.ndarray, t_.Sequence]:
        """
        Args:
            start: The beginning value of the index in the new object. Pass `None` to include everything.
            stop: The ending value of the index in the new object. Pass `None` to include everything.
        Returns:
            A new instance of ICBase with only data from `start` to `stop` in the `index`."""
        wv = np.array(self.index)
        if start is None:
            iStart = None
        else:
            iStart = np.argmin(np.abs(wv - start))
        if stop is None:
            iStop = None
        else:
            iStop = np.argmin(np.abs(wv - stop))
            iStop += 1  # include the end point
            if iStop >= len(wv):  # Include everything
                iStop = None
        data = self.data[:, :, iStart:iStop]
        index = self.index[iStart:iStop]
        return data, index

    def _add(self, other: t_.Union['self.__class__', numbers.Real, np.ndarray]) -> 'self.__class__':  #TODO these don't return the right datatype. They should probably just be gotten rid of
        if isinstance(other, self.__class__):
            if not self._indicesMatch(other):
                raise ValueError(f"{self.__class__} indices are not compatible")
            ret = self.data + other.data
        elif isinstance(other, (numbers.Real, np.ndarray)):
            ret = self.data + other
        else:
            raise NotImplementedError(f"Addition is not supported between {self.__class__} and {type(other)}")
        return ret

    def _sub(self, other: t_.Union['self.__class__', numbers.Real, np.ndarray]) -> 'self.__class__':
        if isinstance(other, self.__class__):
            if not self._indicesMatch(other):
                raise ValueError(f"{self.__class__} indices are not compatible")
            ret = self.data - other.data
        elif isinstance(other, (numbers.Real, np.ndarray)):
            ret = self.data - other
        else:
            raise NotImplementedError(f"Subtraction is not supported between {self.__class__} and {type(other)}")
        return ret

    def _mul(self, other: t_.Union['self.__class__', numbers.Real, np.ndarray]) -> 'self.__class__':
        if isinstance(other, self.__class__):
            if not self._indicesMatch(other):
                raise ValueError(f"{self.__class__} indices are not compatible")
            ret = self.data * other.data
        elif isinstance(other, (numbers.Real, np.ndarray)):
            ret = self.data * other
        else:
            raise NotImplementedError(f"Multiplication is not supported between {self.__class__} and {type(other)}")
        return ret


    def _truediv(self, other: t_.Union['self.__class__', numbers.Real, np.ndarray]) -> 'self.__class__':
        if isinstance(other, self.__class__):
            if not self._indicesMatch(other):
                raise ValueError(f"{self.__class__} indices are not compatible")
            ret = self.data / other.data
        elif isinstance(other, (numbers.Real, np.ndarray)):
            ret = self.data / other
        else:
            raise NotImplementedError(f"Division is not supported between {self.__class__} and {type(other)}")
        return ret

    def __add__(self, other):
        ret = self._add(other)
        new = copy.deepcopy(self)
        new.data = ret
        return new

    def __sub__(self, other):
        ret = self._sub(other)
        new = copy.deepcopy(self)
        new.data = ret
        return new

    def __mul__(self, other):
        ret = self._mul(other)
        new = copy.deepcopy(self)
        new.data = ret
        return new

    __rmul__ = __mul__  # multiplication is commutative. let it work both ways.

    def __truediv__(self, other):
        ret = self._truediv(other)
        new = copy.deepcopy(self)
        new.data = ret
        return new

    def toHdfDataset(self, g: h5py.Group, name: str, fixedPointCompression: bool = True, compression: str = None) -> h5py.Group:
        """
        Save the data of this class to a new HDF dataset.

        Args:
            g (h5py.Group): the parent HDF Group of the new dataset.
            name (str): the name of the new HDF dataset in group `g`.
            fixedPointCompression (bool): if True then save the data in a special 16bit fixed-point format. Testing has shown that this has a
                maximum conversion error of 1.4e-3 percent. Saving is ~10% faster but requires only 50% the hard drive space.
            compression: The value of this argument will be passed to h5py.create_dataset for numpy arrays. See h5py documentation for available options.

        Returns:
            h5py.Group: This is the the same h5py.Group that was passed in a `g`. It should now have a new dataset by the name of 'name'
        """

        if fixedPointCompression:
            # Scale data to span the full range of an unsigned 16bit integer. save as integer and save the min and max
            # needed to scale back to the original data. Testing has shown that this has a maximum conversion error of 1.4e-3 percent.
            # Saving is ~10% faster but requires only 50% the hard drive space. Time can be traded for space by using compression
            # when creating the dataset
            m = self.data.min()
            M = self.data.max()
            fpData = self.data - m
            fpData = fpData / (M - m)
            fpData *= (2 ** 16 - 1)
            fpData = fpData.astype(np.uint16)
            dset = g.create_dataset(name, data=fpData, compression=compression)  # , chunks=(64,64,self.data.shape[2]), compression=2)
            dset.attrs['index'] = np.array(self.index)
            dset.attrs['type'] = np.string_(f"{self._hdfTypeName}_fp")
            dset.attrs['min'] = m
            dset.attrs['max'] = M
        else:
            dset = g.create_dataset(name, data=self.data, compression=compression)
            dset.attrs['index'] = np.array(self.index)
            dset.attrs['type'] = np.string_(self._hdfTypeName)
        return g

    @classmethod
    def decodeHdf(cls, d: h5py.Dataset) -> t_.Tuple[np.array, t_.Tuple[float, ...]]:
        """
        Load a new instance of ICBase from an `h5py.Dataset`

        Args:
            d: The dataset that the ICBase has been saved to

        Returns:
            A tuple containing: (data: The 3D array of `data`,  index: A tuple containing the `index`)
        """
        assert 'type' in d.attrs
        assert 'index' in d.attrs
        if d.attrs['type'].decode() == cls._hdfTypeName: #standard decoding
            return np.array(d), tuple(d.attrs['index'])
        elif d.attrs['type'].decode() == f"{cls._hdfTypeName}_fp": #Fixed point decoding
            M = d.attrs['max']
            m = d.attrs['min']
            arr = np.array(d)
            arr = arr.astype(np.float32) / (2 ** 16 - 1)
            arr *= (M - m)
            arr += m
            return arr, tuple(d.attrs['index'])
        else:
            raise TypeError(f"Got {d.attrs['type'].decode()} instead of {cls._hdfTypeName}")


class ICRawBase(ICBase, ABC):
    """This class represents data cubes which are not derived from other data cubes. They represent raw acquired data that exists as data files on the computer.
    For this reason they may need to have hardware specific corrections applied to them such as normalizing out exposure time, linearizing camera counts,
    subtracting dark counts, etc. The most important change is the addition of `metadata`.

    Args:
        data: A 3-dimensional array containing the data. The dimensions should be [Y, X, Z] where X and Y are the spatial coordinates of the image
            and Z corresponds to the `index` dimension, e.g. wavelength, wavenumber, time, etc.
        index: A tuple containing the values of the index for the data. This could be a tuple of wavelength values, times (in the case of Dynamics), etc.
        metadata: The metadata object associated with this data object.
        processingStatus: An object that keeps track of which processing steps and corrections have been applied to this object.
        dtype (type): the data type that the data should be stored as. The default is numpy.float32.
    """

    @dataclass
    class ProcessingStatus:
        """Keeps track of which processing steps have been applied to an `ICRawBase` object.
        By default none of these things have been done for raw data"""
        cameraCorrected: bool = False
        normalizedByExposure: bool = False
        extraReflectionSubtracted: bool = False
        normalizedByReference: bool = False

        def toDict(self) -> dict:
            return {'camCorrected': self.cameraCorrected, 'exposureNormed': self.normalizedByExposure, 'erSubtracted': self.extraReflectionSubtracted, 'refNormed': self.normalizedByReference}

        @classmethod
        def fromDict(cls, d: dict) -> ICBase.ProcessingStatus:
            return cls(cameraCorrected=d['camCorrected'], normalizedByExposure=d['exposureNormed'], extraReflectionSubtracted=d['erSubtracted'], normalizedByReference=d['refNormed'])

    def __init__(self, data: np.ndarray, index: tuple, metadata: pwsdtmd.MetaDataBase, processingStatus: ProcessingStatus=None, dtype=np.float32):
        super().__init__(data, index, dtype)
        self.metadata = metadata
        if processingStatus:
            self.processingStatus = processingStatus
        else:
            self.processingStatus = ICRawBase.ProcessingStatus(False, False, False, False)

    def normalizeByExposure(self):
        """This is one of the first steps in most analysis pipelines. Data is divided by the camera exposure.
        This way two PwsCube that were acquired at different exposure times will still be on equivalent scales."""
        if not self.processingStatus.cameraCorrected:
            raise Exception(
                "This PwsCube has not yet been corrected for camera effects. are you sure you want to normalize by exposure?")
        if not self.processingStatus.normalizedByExposure:
            self.data = self.data / self.metadata.exposure
        else:
            raise Exception("The PwsCube has already been normalized by exposure.")
        self.processingStatus.normalizedByExposure = True

    def correctCameraEffects(self, correction: _other.CameraCorrection = None, binning: int = None):
        """Subtracts the darkcounts from the data. count is darkcounts per pixel. binning should be specified if
        it wasn't saved in the micromanager metadata. Both method arguments should be able to be loaded automatically
        from the metadata but for older data files they will need to be supplied manually.

        Args:
            correction: The cameracorrection object providing information on how to correct the data.
            binning: The binning that the raw data was imaged at. 2 = 2x2 binning, 3 = 3x3 binning, etc.
        """
        if self.processingStatus.cameraCorrected:
            raise Exception("This PwsCube has already had it's camera correction applied!")
        if binning is None:
            binning = self.metadata.binning
            if binning is None: raise ValueError('Binning metadata not found. Binning must be specified in function argument.')
        if correction is None:
            correction = self.metadata.cameraCorrection
            if correction is None: raise ValueError('other.CameraCorrection metadata not found. Binning must be specified in function argument.')
        count = correction.darkCounts * binning ** 2  # Account for the fact that binning multiplies the darkcount.
        self.data = self.data - count
        if correction.linearityPolynomial is None or correction.linearityPolynomial == (1.0,):
            pass
        else:
            self.data = np.polynomial.polynomial.polyval(self.data, (0.0,) + correction.linearityPolynomial)  # The [0] item is the y-intercept (already handled by the darkcount)
        self.processingStatus.cameraCorrected = True
        return

    @abstractmethod
    def normalizeByReference(self, reference: 'self.__class__'):
        """Normalize the raw data of this data cube by a reference cube to result in data representing
        arbitrarily scaled reflectance.

        Args:
            reference: A reference acquisition. Usually an image taken from a blank piece of glass.
        """
        pass

    @abstractmethod
    def subtractExtraReflection(self, extraReflection: ExtraReflectionCube):
        """
        Subtract the portion of the signal that is due to internal reflections of the optical system from the data.

        Args:
            extraReflection: A calculated data cube indicating in units of camera counts how much of the data is from
                unwanted internal reflections of the system.
        """
        pass

    def performFullPreProcessing(self, reference: 'self.__class__', referenceMaterial: Material, extraReflectance: ExtraReflectanceCube, cameraCorrection: t_.Optional[_other.CameraCorrection] = None):
        """
        Use the `subtractExtraReflection`, `normalizeByReference`, `correctCameraEffects`, and `normalizeByExposure`
        methods to perform the standard pre-processing that is done before analysis.

        Note: This will also end up applying corrections to the reference data. If you want to perform pre-processing on a whole batch
        of data then you should implement your own script based on what is done here.

        Args:
            reference: A data cube to be used as a reference for normalization. Usually an image of a blank dish with cell media or air.
            referenceMaterial: The material that was imaged in the reference dish. The theoretically expected reflectance will be calculated
                assuming a "Glass/{Material}" reflective interface.
            extraReflectance: The data cube containing system internal reflectance calibration information about the specific system
                configuration that the data was taken with.
        """
        from pwspy.utility.reflection.reflectanceHelper import getReflectance

        assert not reference.processingStatus.extraReflectionSubtracted
        assert not reference.processingStatus.cameraCorrected
        assert not reference.processingStatus.normalizedByExposure
        reference.correctCameraEffects(cameraCorrection)
        reference.normalizeByExposure()
        self.correctCameraEffects(cameraCorrection)
        self.normalizeByExposure()

        reflection = ExtraReflectionCube.create(extraReflectance, getReflectance(Material.Glass, referenceMaterial), reference)
        reference.subtractExtraReflection(reflection)
        self.subtractExtraReflection(reflection)

        self.normalizeByReference(reference)

    @staticmethod
    @abstractmethod
    def getMetadataClass() -> t_.Type[pwsdtmd.MetaDataBase]:
        """

        Returns:
            The metadata class associated with this subclass of ICRawBase
        """
        pass

    def toHdfDataset(self, g: h5py.Group, name: str, fixedPointCompression: bool = True) -> h5py.Group:
        """
        Save this object into an HDF dataset.

        Args:
            g: The `h5py.Group` object to create a new dataset in.
            name: The name of the new dataset.
            fixedPointCompression: If True then the data will be converted from floating point to 16-bit fixed point.
                This results in approximately half the storage requirements at a very slight loss in precision.

        Returns:
            A reference to the `h5py.Group` passed in as `g`.

        """
        g = ICBase.toHdfDataset(self, g, name, fixedPointCompression)
        self.metadata.encodeHdfMetadata(g[name])
        g[name].attrs['processingStatus'] = np.string_(json.dumps(self.processingStatus.toDict()))
        return g

    @classmethod
    def decodeHdf(cls, d: h5py.Dataset) -> t_.Tuple[np.array, t_.Tuple[float, ...], dict, ProcessingStatus]:
        """
        Load a new instance of ICRawBase from an `h5py.Dataset`

        Args:
            d: The dataset that the ICBase has been saved to

        Returns:
            A tuple containing:
                data: The 3D array of `data`
                index: A tuple containing the `index`
                metadata: A dictionary containing metadata.
                procStatus: The processing status of the object.
        """
        arr, index = super().decodeHdf(d)
        mdDict = cls.getMetadataClass().decodeHdfMetadata(d)
        if 'processingStatus' in d.attrs:
            processingStatus = cls.ProcessingStatus.fromDict(json.loads(d.attrs['processingStatus']))
        else:  # Some old hdf files won't have this field, that's ok.
            processingStatus = None
        return arr, index, mdDict, processingStatus


class DynCube(ICRawBase):
    """A class representing a single acquisition of PWS Dynamics. In which the wavelength is held constant and the 3rd
    dimension of the data is time rather than wavelength. This can be analyzed to reveal information about diffusion rate.
    Contains methods for loading and saving to multiple formats as well as common operations used in analysis.

    Args:
        data: A 3-dimensional array containing the data. The dimensions should be [Y, X, Z] where X and Y are the spatial coordinates of the image
            and Z corresponds to the `index` dimension, e.g. wavelength, wavenumber, time, etc.
        metadata: The metadata object associated with this data object.
        processingStatus: An object that keeps track of which processing steps and corrections have been applied to this object.
        dtype: the data type that the data should be stored as. The default is numpy.float32.
    """

    _hdfTypeName = "DynCube"  # This is used for saving/loading from HDF. Important not to change it or old files will stop working.

    def __init__(self, data, metadata: pwsdtmd.DynMetaData, processingStatus: ICRawBase.ProcessingStatus = None, dtype=np.float32):
        assert isinstance(metadata, pwsdtmd.DynMetaData)
        super().__init__(data, metadata.times, metadata, processingStatus=processingStatus, dtype=dtype)

    @staticmethod
    def getMetadataClass() -> t_.Type[pwsdtmd.DynMetaData]:
        return pwsdtmd.DynMetaData

    @property
    def times(self) -> t_.Tuple[float, ...]:
        """Unlike PWS where we operate along the dimension of wavelength, in dynamics we operate along the dimension of time.

        Returns:
            A tuple of the time values for each 2d slice along the 3rd axis of the `data` array.
        """
        return self.index

    @classmethod
    def fromMetadata(cls, meta: pwsdtmd.DynMetaData, lock: mp.Lock = None) -> DynCube:
        """
        Load a new instance of `DynCube` based on the information contained in a `DynMetaData` object.

        Args:
            meta: The metadata object to be used for loading.
            lock: An optional `Lock` used to synchronize IO operations in multithreaded and multiprocessing applications.

        Returns:
            A new instance of `DynCube`.
        """
        if meta.fileFormat == pwsdtmd.DynMetaData.FileFormats.Tiff:
            return cls.fromTiff(meta.filePath, metadata=meta, lock=lock)
        elif meta.fileFormat == pwsdtmd.DynMetaData.FileFormats.RawBinary:
            return cls.fromOldPWS(meta.filePath, metadata=meta, lock=lock)
        elif meta.fileFormat is None:
            return cls.loadAny(meta.filePath, metadata=meta, lock=lock)
        else:
            raise TypeError("Invalid FileFormat")

    @classmethod
    def loadAny(cls, directory: str, metadata: t_.Optional[pwsdtmd.DynMetaData] = None, lock: t_.Optional[mp.Lock] = None) -> DynCube:
        """
        Attempt to load a `DynCube` for any format of file in `directory`

        Args:
            directory: The directory containing the data files.
            metadata: The metadata object associated with this acquisition
            lock: A `Lock` object used to synchronized IO in multithreading and multiprocessing applications.

        Returns:
            A new instance of `DynCube`.
        """
        try:
            return DynCube.fromTiff(directory, metadata=metadata, lock=lock)
        except:
            try:
                return DynCube.fromOldPWS(directory, metadata=metadata, lock=lock)
            except:
                raise OSError(f"Could not find a valid PWS image cube file at {directory}.")

    @classmethod
    def fromOldPWS(cls, directory, metadata: pwsdtmd.DynMetaData = None,  lock: mp.Lock = None) -> DynCube:
        """Loads from the file format that was saved by the all-matlab version of the Basis acquisition code.
        Data was saved in raw binary to a file called `image_cube`. Some metadata was saved to .mat files called
        `info2` and `info3`.

        Args:
            directory: The directory containing the data files.
            metadata: The metadata object associated with this acquisition
            lock: A `Lock` object used to synchronized IO in multithreading and multiprocessing applications.

        Returns:
            A new instance of `DynCube`.
        """
        if lock is not None:
            lock.acquire()
        try:
            if metadata is None:
                metadata = pwsdtmd.DynMetaData.fromOldPWS(directory)
            with open(os.path.join(directory, 'image_cube'), 'rb') as f:
                data = np.frombuffer(f.read(), dtype=np.uint16)
            data = data.reshape((metadata.dict['imgHeight'], metadata.dict['imgWidth'], len(metadata.times)), order='F')
        finally:
            if lock is not None:
                lock.release()
        data = data.copy(order='C')
        return cls(data, metadata)

    @classmethod
    def fromTiff(cls, directory, metadata: pwsdtmd.DynMetaData = None, lock: mp.Lock = None) -> DynCube:
        """Load a dyanmics acquisition from a tiff file. if the metadata for the acquisition has already been loaded then you can provide
        is as the `metadata` argument to avoid loading it again. the `lock` argument is an optional place to provide a multiprocessing.Lock
        which can be used when multiple files in parallel to avoid giving the hard drive too many simultaneous requests, this is probably not necessary.

        Args:
            directory: The directory containing the data files.
            metadata: The metadata object associated with this acquisition
            lock: A `Lock` object used to synchronized IO in multithreading and multiprocessing applications.

        Returns:
            A new instance of `DynCube`.
        """
        if lock is not None:
            lock.acquire()
        try:
            if metadata is None:
                metadata = pwsdtmd.DynMetaData.fromTiff(directory)
            if os.path.exists(os.path.join(directory, 'dyn.tif')):
                path = os.path.join(directory, 'dyn.tif')
            else:
                raise OSError("No Tiff file was found at:", directory)
            with tf.TiffFile(path) as tif:
                data = np.rollaxis(tif.asarray(), 0, 3)  # Swap axes to match y,x,lambda convention.
        finally:
            if lock is not None:
                lock.release()
        data = data.copy(order='C')
        return cls(data, metadata)

    def normalizeByReference(self, reference: t_.Union[DynCube, np.ndarray]):
        """This method can accept either a DynCube (in which case it's average over time will be calculated and used for
        normalization) or a 2d numpy Array which should represent the average over time of a reference DynCube. The array
        should be 2D and its shape should match the first two dimensions of this DynCube.

        Args:
            reference: Reference data for normalization. Usually an image of a blank piece of glass.
        """
        logger = logging.getLogger(__name__)
        if self.processingStatus.normalizedByReference:
            raise Exception("This cube has already been normalized by a reference.")
        if not self.processingStatus.cameraCorrected:
            logger.warning("This cube has not been corrected for camera effects. This is highly reccomended before performing any analysis steps.")
        if not self.processingStatus.normalizedByExposure:
            logger.warning("This cube has not been normalized by exposure. This is highly reccomended before performing any analysis steps.")
        if isinstance(reference, DynCube):
            if not reference.processingStatus.cameraCorrected:
                logger.warning("The reference cube has not been corrected for camera effects. This is highly reccomended before performing any analysis steps.")
            if not reference.processingStatus.normalizedByExposure:
                logger.warning("The reference cube has not been normalized by exposure. This is highly reccomended before performing any analysis steps.")
            mean = reference.data.mean(axis=2)
        elif isinstance(reference, np.ndarray):
            assert len(reference.shape) == 2
            assert reference.shape[0] == self.data.shape[0]
            assert reference.shape[1] == self.data.shape[1]
            mean = reference
        else:
            raise TypeError(f"`reference` must be either DynCube or numpy.ndarray, not {type(reference)}")
        self.data = self.data / mean[:, :, None]
        self.processingStatus.normalizedByReference = True

    def subtractExtraReflection(self, extraReflection: np.ndarray): # Inherit docstring
        assert self.data.shape[:2] == extraReflection.shape
        if not self.processingStatus.normalizedByExposure:
            raise Exception("This DynCube has not yet been normalized by exposure. are you sure you want to normalize by exposure?")
        if not self.processingStatus.extraReflectionSubtracted:
            self.data = self.data - extraReflection[:, :, None]
            self.processingStatus.extraReflectionSubtracted = True
        else:
            raise Exception("The DynCube has already has extra reflection subtracted.")

    def selIndex(self, start, stop) -> DynCube: # Inherit docstring
        data, index = super().selIndex(start, stop)
        md = self.metadata
        md.dict['times'] = index
        return DynCube(data, md)

    def getAutocorrelation(self) -> np.ndarray:
        """
        Returns the autocorrelation function of dynamics data along the time axis. The ACF is calculated using
        fourier transforms using IFFT(FFT(data)*conj(FFT(data)))/length(data).

        Returns:
            A 3D array of the autocorrelation function of the original data.
        """
        data = self.data - self.data.mean(axis=2)[:, :, None]  # By subtracting the mean we get an ACF where the 0-lag value is the variance of the signal.
        F = np.fft.rfft(data, axis=2)
        ac = np.fft.irfft(F * np.conjugate(F), axis=2) / data.shape[2]
        return ac

    def filterDust(self, kernelRadius: float, pixelSize: float = None):
        """
        This method blurs the data of the cube along the X and Y dimensions. This is useful if the cube is being
        used as a reference to normalize other cube. It helps blur out dust and other unwanted small features.

        Args:
            kernelRadius: The `sigma` of the gaussian kernel used for blurring. A greater value results in greater
                blurring. If `pixelSize` is provided then this is in units of `pixelSize`, otherwise it is in units of
                pixels.
            pixelSize: The size (usualy in units of microns) of each pixel in the datacube. This can generally be loaded
                automatically from the metadata.
        """
        if pixelSize is None:
            pixelSize = self.metadata.pixelSizeUm
            if pixelSize is None:
                raise ValueError("DynCube Metadata does not have a `pixelSizeUm` saved. please manually specify pixel size. use pixelSize=1 to make `kernelRadius in units of pixels.")
        super().filterDust(kernelRadius, pixelSize)

    @classmethod
    def fromHdfDataset(cls, d: h5py.Dataset):  # Inherit docstring
        data, index, mdDict, processingStatus = cls.decodeHdf(d)
        md = pwsdtmd.DynMetaData(mdDict, fileFormat=pwsdtmd.DynMetaData.FileFormats.Hdf)
        return cls(data, md, processingStatus=processingStatus)


class ExtraReflectanceCube(ICBase):
    """This class represents a 3D data cube of the extra reflectance in a PWS system.

    Args:
        data: A 3D array of the extra reflectance in the system. It's values are in units of reflectance (between 0 and 1).
        wavelengths: The wavelengths associated with each 2D slice along the 3rd axis of the data array.
        metadata: Metadata
    Attributes:
        metadata (ERMetaData): metadata
        data (ndarray): data
    """

    _hdfTypeName = "ExtraReflectanceCube"  # This is used for saving/loading from HDF. Important not to change it or old files will stop working.

    def __init__(self, data: np.ndarray, wavelengths: t_.Tuple[float, ...], metadata: pwsdtmd.ERMetaData):
        assert isinstance(metadata, pwsdtmd.ERMetaData)
        if data.max() > 1 or data.min() < 0:
            logging.getLogger(__name__).warning("Reflectance values must be between 0 and 1")
        self.metadata = metadata
        ICBase.__init__(self, data, wavelengths)

    @property
    def wavelengths(self) -> t_.Tuple[float, ...]:
        """

        Returns:
            The wavelengths corresponding to each element along the 3rd axis of `self.data`.
        """
        return self.index

    @classmethod
    def fromHdfFile(cls, directory: str, name: str) -> ExtraReflectanceCube:
        """
        Load an ExtraReflectanceCube from an HDF5 file. `name` should be the file name, excluding the '_ExtraReflectance.h5' suffix.

        Args:
            directory: The path to the folder containing the HDF file.
            name: The `name` that the cube was saved as.
        Returns:
            A new instance of `ExtraReflectanceCube` loaded from HDF.
        """
        filePath = pwsdtmd.ERMetaData.dirName2Directory(directory, name)
        with h5py.File(filePath, 'r') as hf:
            dset = hf[pwsdtmd.ERMetaData.DATASETTAG]
            return cls.fromHdfDataset(dset, filePath=filePath)

    def toHdfFile(self, directory: str, name: str):
        """
        Save an ExtraReflectanceCube to an HDF5 file. The filename will be `name` with the '_ExtraReflectance.h5' suffix.

        Args:
            directory: The path to the folder to save the HDF file to.
            name: The `name` that the cube should be saved as.
        """
        savePath = pwsdtmd.ERMetaData.dirName2Directory(directory, name)
        if os.path.exists(savePath):
            raise OSError(f"The path {savePath} already exists.")
        with h5py.File(savePath, 'w') as hf:
            self.toHdfDataset(hf)

    def toHdfDataset(self, g: h5py.Group) -> h5py.Group:
        """
        Save the ExtraReflectanceCube to an HDF5 dataset. `g` should be an h5py Group or File.

        Args:
            g: The `h5py.Group` to save to.
        """
        g = super().toHdfDataset(g, pwsdtmd.ERMetaData.DATASETTAG)
        g = self.metadata.toHdfDataset(g)
        return g

    @classmethod
    def fromHdfDataset(cls, d: h5py.Dataset, filePath: str = None) -> ExtraReflectanceCube:
        """Load the ExtraReflectanceCube from `d`, an HDF5 dataset.

        Args:
            d: The `h5py.Dataset` to load the cube from.
            filePath: The path to the HDF file that the dataset came from.
        Returns:
            A new instance of `ExtraReflectanceCube` loaded from HDF.
        """
        data, index = cls.decodeHdf(d)
        md = pwsdtmd.ERMetaData.fromHdfDataset(d, filePath=filePath)
        return cls(data, index, md)

    @classmethod
    def fromMetadata(cls, md: pwsdtmd.ERMetaData):
        """Load an ExtraReflectanceCube from an ERMetaData object corresponding to an HDF5 file.

        Args:
            md: The metadata to be used for loading the data file.
        """
        directory, name = pwsdtmd.ERMetaData.directory2dirName(md.filePath)
        return cls.fromHdfFile(directory, name)


class ExtraReflectionCube(ICBase):
    """This class is meant to be constructed from an ExtraReflectanceCube along with additional reference measurement
    information. Rather than being in units of reflectance (between 0 and 1) it is in the same units as the reference measurement
    that is provided with, usually counts/ms or just counts.

    Args:
        data: The 3D array of the extra reflection in the system. In units of counts/ms or just counts
        wavelengths: The wavelengths associated with each 2D slice along the 3rd axis of the data array.
        metadata: Metadata
    """

    _hdfTypeName = "ExtraReflectionCube"  # This is used for saving/loading from HDF. Important not to change it or old files will stop working.

    def __init__(self, data: np.ndarray, wavelengths: t_.Tuple[float, ...], metadata: pwsdtmd.ERMetaData):
        super().__init__(data, wavelengths)
        self.metadata = metadata

    @classmethod
    def create(cls, reflectance: ExtraReflectanceCube, theoryR: pd.Series, reference: PwsCube) -> ExtraReflectionCube:
        """
        Construct and ExtraReflectionCube from an ExtraReflectanceCube and a reference measurement. The resulting
        ExtraReflectionCube will be in the same units as `reference`. `theoryR` should be a spectrum describing the theoretically
        expected reflectance of the reference data cube. Both `theoryR` and `reflectance` should be in units of reflectance
        (between 0 and 1).

        Args:
            reflectance: The `ExtraReflectanceCube` to construct an `ExtraReflectionCube` from.
            theoryR: The theoretically predicted reflectance of material imaged in `reference`.
            reference: A PWS image of a blank glass-{material} interface, usually water.
        Returns:
            A new instance of `ExtraReflectionCube`.
        """
        I0 = reference.data / (np.array(theoryR)[None, None, :] + reflectance.data)  # I0 is the intensity of the illumination source, reconstructed in units of `counts`. this is an inversion of our assumption that reference = I0*(referenceReflectance + extraReflectance)
        data = reflectance.data * I0  # converting extraReflectance to the extra reflection in units of counts
        return cls(data, reflectance.wavelengths, reflectance.metadata)


class PwsCube(ICRawBase):
    """
    A class representing a single PWS acquisition. Contains methods for loading and saving to multiple formats as
    well as common operations used in analysis.

    Args:
        data: A 3-dimensional array containing the data. The dimensions should be [Y, X, Z] where X and Y are the spatial coordinates of the image
            and Z corresponds to the `index` dimension, e.g. wavelength, wavenumber, time, etc.
        metadata: The metadata object associated with this data object.
        processingStatus: An object that keeps track of which processing steps and corrections have been applied to this object.
        dtype (type): the data type that the data should be stored as. The default is numpy.float32.
    """

    _hdfTypeName = "ImCube"  # This is used for saving/loading from HDF. Important not to change it or old files will stop working.

    def __init__(self, data, metadata: pwsdtmd.PwsMetaData, processingStatus: ICRawBase.ProcessingStatus = None, dtype=np.float32):
        assert isinstance(metadata, pwsdtmd.PwsMetaData)
        super().__init__(data, metadata.wavelengths, metadata, processingStatus=processingStatus, dtype=dtype)

    @property
    def wavelengths(self):
        """
        A tuple containing the values of the wavelengths for the data.
        """
        return self.index

    @classmethod
    def loadAny(cls, directory: str, metadata: pwsdtmd.PwsMetaData = None, lock: mp.Lock = None):
        """
        Attempt to load a `PwsCube` for any format of file in `directory`

        Args:
            directory: The directory containing the data files.
            metadata: The metadata object associated with this acquisition
            lock: A `Lock` object used to synchronized IO in multithreading and multiprocessing applications.

        Returns:
            A new instance of `PwsCube`.
        """
        try:
            return PwsCube.fromTiff(directory, metadata=metadata, lock=lock)
        except:
            try:
                return PwsCube.fromOldPWS(directory, metadata=metadata, lock=lock)
            except:
                try:
                    return PwsCube.fromNano(directory, metadata=metadata, lock=lock)
                except:
                    raise OSError(f"Could not find a valid PWS image cube file at {directory}.")

    @classmethod
    def fromOldPWS(cls, directory: str, metadata: pwsdtmd.PwsMetaData = None, lock: mp.Lock = None):
        """
        Loads from the file format that was saved by the all-matlab version of the Basis acquisition code.
        Data was saved in raw binary to a file called `image_cube`. Some metadata was saved to .mat files called
        `info2` and `info3`.

        Args:
            directory: The directory containing the data files.
            metadata: The metadata object associated with this acquisition
            lock: A `Lock` object used to synchronized IO in multithreading and multiprocessing applications.

        Returns:
            A new instance of `PwsCube`.
        """
        if lock is not None:
            lock.acquire()
        try:
            if metadata is None:
                metadata = pwsdtmd.PwsMetaData.fromOldPWS(directory)
            with open(os.path.join(directory, 'image_cube'), 'rb') as f:
                data = np.frombuffer(f.read(), dtype=np.uint16)
            data = data.reshape((metadata.dict['imgHeight'], metadata.dict['imgWidth'], len(metadata.wavelengths)), order='F')
        finally:
            if lock is not None:
                lock.release()
        data = data.copy(order='C')
        return cls(data, metadata)

    @classmethod
    def fromTiff(cls, directory, metadata: pwsdtmd.PwsMetaData = None, lock: mp.Lock = None):
        """
        Loads from a 3D tiff file named `pws.tif`, or in some older data `MMStack.ome.tif`. Metadata can be stored in
        the tags of the tiff file but if there is a pwsmetadata.json file found then this is preferred.
        A multiprocessing.Lock object can be passed to this function so that it will acquire a lock during the
        hard-drive intensive parts of the function. this is useful in multi-core contexts.

        Args:
            directory: The directory containing the data files.
            metadata: The metadata object associated with this acquisition
            lock: A `Lock` object used to synchronized IO in multithreading and multiprocessing applications.

        Returns:
            A new instance of `PwsCube`.
        """
        if lock is not None:
            lock.acquire()
        try:
            if metadata is None:
                metadata = pwsdtmd.PwsMetaData.fromTiff(directory)
            if os.path.exists(os.path.join(directory, 'MMStack.ome.tif')):
                path = os.path.join(directory, 'MMStack.ome.tif')
            elif os.path.exists(os.path.join(directory, 'pws.tif')):
                path = os.path.join(directory, 'pws.tif')
            else:
                raise OSError("No Tiff file was found at:", directory)
            with tf.TiffFile(path) as tif:
                data = np.rollaxis(tif.asarray(), 0, 3)  # Swap axes to match y,x,lambda convention.
        finally:
            if lock is not None:
                lock.release()
        data = data.copy(order='C')
        return cls(data, metadata)

    @classmethod
    def fromNano(cls, directory: str, metadata: pwsdtmd.PwsMetaData = None, lock: mp.Lock = None) -> PwsCube:
        """
        Loads from the file format used at NC. all data and metadata is contained in a .mat file.

        Args:
            directory: The directory containing the data files.
            metadata: The metadata object associated with this acquisition
            lock: A `Lock` object used to synchronized IO in multithreading and multiprocessing applications.

        Returns:
            A new instance of `PwsCube`.
        """
        path = os.path.join(directory, 'imageCube.mat')
        if lock is not None:
            lock.acquire()
        try:
            if metadata is None:
                metadata = pwsdtmd.PwsMetaData.fromNano(directory)
            with h5py.File(path, 'r') as hf:
                data = np.array(hf['imageCube'])
                data = data.transpose((2, 1, 0))  # Re-order axes to match the shape of ROIs and thumbnails.
                data = data.copy(order='C')  # Copying the data to 'C' order has shows to provide some benefit for performance in selecting out spectra.
        finally:
            if lock is not None:
                lock.release()
        return cls(data, metadata)

    @classmethod
    def fromMetadata(cls, meta: pwsdtmd.PwsMetaData, lock: mp.Lock = None) -> PwsCube:
        """
        If provided with an PwsMetaData object this function will automatically select the correct file loading method
        and will return the associated PwsCube.

        Args:
            meta: The metadata to use to load the object from.
            lock: A `Lock` object used to synchronized IO in multithreading and multiprocessing applications.

        Returns:
            A new instance of `PwsCube`.
        """
        assert isinstance(meta, pwsdtmd.PwsMetaData)
        if meta.fileFormat == pwsdtmd.PwsMetaData.FileFormats.Tiff:
            return cls.fromTiff(meta.filePath, metadata=meta, lock=lock)
        elif meta.fileFormat == pwsdtmd.PwsMetaData.FileFormats.RawBinary:
            return cls.fromOldPWS(meta.filePath, metadata=meta, lock=lock)
        elif meta.fileFormat == pwsdtmd.PwsMetaData.FileFormats.NanoMat:
            return cls.fromNano(meta.filePath, metadata=meta, lock=lock)
        elif meta.fileFormat is None:
            return cls.loadAny(meta.filePath, metadata=meta, lock=lock)
        else:
            raise TypeError("Invalid FileFormat")

    def toOldPWS(self, directory):
        """
        Save this object to the old .mat based storage format.

        Args:
            directory: The path to the folder to save the data files to.
        """
        if os.path.exists(directory):
            raise OSError("The specified directory already exists.")
        os.mkdir(directory)
        m = self.metadata
        try:
            systemId = m.dict['systemId']
        except KeyError:
            systemId = 0
        info2 = {'info2': np.array([m.wavelengths[0], m.wavelengths[1]-m.wavelengths[0], m.wavelengths[-1], m.exposure, 0, 0, 0, 0, 0, 0],
                                   dtype=np.float64)}
        info3 = {'info3': np.array(
            [systemId, m.exposure, self.data.shape[0], self.data.shape[1], 1970, 1, 1, 0, 0, 0, 0, 0], #Use data 1/1/1970 since we don't have a real acquisition date.
            dtype=np.float64)}  # The new way
        wv = {"WV": m.wavelengths}
        savemat(os.path.join(directory, 'info2'), info2)
        savemat(os.path.join(directory, 'info3'), info3)
        savemat(os.path.join(directory, 'WV'), wv)
        self._saveThumbnail(directory)
        with open(os.path.join(directory, 'image_cube'), 'wb') as f:
            f.write(self.data.astype(np.uint16).tobytes(order='F'))

    def _saveThumbnail(self, directory):
        """Used to save a thumbnail image called `image_bd.tif` this is useful for quickly viewing the cells without
        having to load and process all the data."""
        im = self.data[:, :, self.data.shape[-1] // 2]
        normedIm = im - np.percentile(im, 0.01)  # .01 percent saturation
        normedIm[normedIm<0] = 0 #Don't allow negative values.
        normedIm = normedIm / np.percentile(normedIm, 99.99)
        normedIm[normedIm>1] = 1 #Keep eveything below 1
        normedIm = (normedIm * 255).astype(np.uint8)
        im = tf.TiffWriter(os.path.join(directory, 'image_bd.tif'))
        im.save(normedIm)
        im.close()

    def toTiff(self, outpath: str, dtype=np.uint16):
        """Save the PwsCube to the standard TIFF file format.

        Args:
            outpath: The path to save the new TIFF file to.
        """
        im = self.data
        im = im.astype(dtype)
        os.mkdir(outpath)
        self._saveThumbnail(outpath)
        with tf.TiffWriter(open(os.path.join(outpath, 'pws.tif'), 'wb')) as w:
            w.save(np.rollaxis(im, -1, 0), metadata=self.metadata.dict)
        self.metadata.metadataToJson(outpath)

    def selIndex(self, start: t_.Optional[float], stop: t_.Optional[float]) -> PwsCube:  # Inherit docstring
        data, index = super().selIndex(start, stop)
        md = copy.deepcopy(self.metadata)  # We are creating a copy of the metadata object because modifying the original metadata object can cause weird issues.
        assert md.dict is not self.metadata.dict
        md.dict['wavelengths'] = index
        return PwsCube(data, md)

    @staticmethod
    def getMetadataClass() -> t_.Type[pwsdtmd.PwsMetaData]:  # Inherit docstring
        return pwsdtmd.PwsMetaData

    @classmethod
    def fromHdfDataset(cls, d: h5py.Dataset):
        """Load an PwsCube from an HDF5 dataset."""
        data, index, mdDict, processingStatus = cls.decodeHdf(d)
        md = pwsdtmd.PwsMetaData(mdDict, fileFormat=pwsdtmd.PwsMetaData.FileFormats.Hdf)
        return cls(data, md, processingStatus=processingStatus)

    def filterDust(self, kernelRadius: float, pixelSize: float = None) -> None:
        """This method blurs the data of the PwsCube along the X and Y dimensions. This is useful if the PwsCube is being
        used as a reference to normalize other PwsCube. It helps blur out dust adn other unwanted small features.

        Args:
            kernelRadius: The `sigma` of the gaussian kernel used for blurring. A greater value results in greater
                blurring. If `pixelSize` is provided then this is in units of `pixelSize`, otherwise it is in units of
                pixels.
            pixelSize: The size (usualy in units of microns) of each pixel in the datacube. This can generally be loaded
                automatically from the metadata.
        """
        if pixelSize is None:
            pixelSize = self.metadata.pixelSizeUm
            if pixelSize is None:
                raise ValueError("PwsCube Metadata does not have a `pixelSizeUm` saved. please manually specify pixel size. use pixelSize=1 to make `kernelRadius in units of pixels.")
        super().filterDust(kernelRadius, pixelSize)

    def normalizeByReference(self, reference: PwsCube):
        """Normalize the raw data of this data cube by a reference cube to result in data representing
        arbitrarily scaled reflectance.

        Args:
            reference (PwsCube): A reference acquisition (Usually a blank spot on a dish). The data of this acquisition will be divided by the data of the reference
        """
        logger = logging.getLogger(__name__)
        if self.processingStatus.normalizedByReference:
            raise Exception("This PwsCube has already been normalized by a reference.")
        if not self.processingStatus.cameraCorrected:
            logger.warning("This PwsCube has not been corrected for camera effects. This is highly reccomended before performing any analysis steps.")
        if not self.processingStatus.normalizedByExposure:
            logger.warning("This PwsCube has not been normalized by exposure. This is highly reccomended before performing any analysis steps.")
        if not reference.processingStatus.cameraCorrected:
            logger.warning("The reference PwsCube has not been corrected for camera effects. This is highly reccomended before performing any analysis steps.")
        if not reference.processingStatus.normalizedByExposure:
            logger.warning("The reference PwsCube has not been normalized by exposure. This is highly reccomended before performing any analysis steps.")
        self.data = self.data / reference.data
        self.processingStatus.normalizedByReference = True

    def subtractExtraReflection(self, extraReflection: ExtraReflectionCube):  # Inherit docstring
        assert self.data.shape == extraReflection.data.shape
        if not self.processingStatus.normalizedByExposure:
            raise Exception("This PwsCube has not yet been normalized by exposure. Are you sure you want to subtract system reflectance before doing this?")
        if not self.processingStatus.extraReflectionSubtracted:
            self.data = self.data - extraReflection.data
            self.processingStatus.extraReflectionSubtracted = True
        else:
            raise Exception("The PwsCube has already has extra reflection subtracted.")

    def __repr__(self):
        return f"{self.__class__.__name__}(id={self.metadata.idTag})"


class KCube(ICBase):
    """A class representing an PwsCube after being transformed from being described in terms of wavelength to
    wavenumber (k-space). Much of the analysis operated in terms of k-space.

    Args:
        data: A 3-dimensional array containing the data. The dimensions should be [Y, X, Z] where X and Y are the spatial coordinates of the image
            and Z corresponds to the `index` dimension, e.g. wavelength, wavenumber, time, etc.
        wavenumbers: A sequence indicating the wavenumber associated with each 2D slice along the 3rd axis of the `data` array.
        metadata: The metadata object associated with this data object.
    """

    _hdfTypeName = "KCube"  # This is used for saving/loading from HDF. Important not to change it or old files will stop working.

    def __init__(self, data: np.ndarray, wavenumbers: t_.Tuple[float], metadata: pwsdtmd.PwsMetaData = None):
        self.metadata = metadata #Just saving a reference to the original PwsCube in case we want to reference it.
        ICBase.__init__(self, data, wavenumbers, dtype=np.float32)

    @classmethod
    def fromPwsCube(cls, cube: PwsCube) -> KCube:
        """
        Convert an PwsCube into a KCube. Data is converted from wavelength to wavenumber (1/lambda), interpolation is
        then used to linearize the data in terms of wavenumber.

        Args:
            cube: The `PwsCube` object to generate a `KCube` object from.

        Returns:
            A new instance of `KCube`
        """
        # Convert to wavenumber and reverse the order so we are ascending in order. Units of radian/micron
        wavenumbers = (2 * np.pi) / (np.array(cube.wavelengths, dtype=np.float64) * 1e-3)[::-1]
        data = cube.data[:, :, ::-1]
        # Generate evenly spaced wavenumbers
        #        dk = (self.wavenumbers[-1] - self.wavenumbers[0])/(len(self.wavenumbers)-1);
        evenWavenumbers = np.linspace(wavenumbers[0], wavenumbers[-1], num=len(wavenumbers), dtype=np.float64)
        # Interpolate to the evenly spaced wavenumbers
        interpFunc = spi.interp1d(wavenumbers, data, kind='linear', axis=2)
        data = interpFunc(evenWavenumbers)
        return cls(data, tuple(evenWavenumbers.astype(np.float32)), metadata=cube.metadata)

    @property
    def wavenumbers(self) -> t_.Tuple[float, ...]:
        return self.index

    def getOpd(self, useHannWindow: bool, indexOpdStop: int = None, mask: np.ndarray = None) -> t_.Tuple[np.ndarray, np.ndarray]:
        """
        Calculate the Fourier transform of each spectra. This can be used to get the distance (in terms of OPD) to
        objects that are reflecting light.

        Args:
            useHannWindow: If True, apply a Hann window to the data before the FFT. This reduces spectral resolution but
                improves dynamic range and reduces "frequency leakage".
            indexOpdStop: This parameter is a holdover from the original MATLAB implementation. Truncates the 3rd axis
                of the OPD array.
            mask: A 2D boolean numpy array indicating which pixels should be processed.

        Returns:
            A tuple containing: `opd`: The 3D array of values, `opdIndex`: The sequence of OPD values associated with each
                2D slice along the 3rd axis of the `opd` data.

        """
        opd = _FFTHelper.getFFTMagnitude(self.data, useHannWindow, normalization=_FFTHelper.Normalization.POWER)
        fftSize = opd.shape[-1]  # Due to FFT interpolation the FFT will be longer than the original data.

        # Isolate the desired values in the OPD.
        opd = opd[:, :, :indexOpdStop]

        if mask is not None:  # Get the average opd within the ROI of the bool array `mask`
            opd = opd[mask].mean(axis=0)

        # Generate the opd values for the current OPD.
        dk = self.wavenumbers[1] - self.wavenumbers[0]  # The interval that our linear array of wavenumbers is spaced by. Units: radians / micron
        maxFrequency = 1 / (dk * 2)  # Max resolvable frequency is half of the sampling frequency. Units: microns / radian
        maxOpd = 2 * np.pi * maxFrequency  # Convert to units of microns. This is the maximum OPD value we can get with our sample interval. tighter wavenumber spacing increases OPD range. units of microns
        opdVals = np.linspace(0, maxOpd, num=fftSize)

        # Above is how the old MATLAB code calculated the frequencies. IMO the code below is simpler and more understandable but we'll stick with the old code.
        # opdVals = 2 * np.pi * np.fft.rfftfreq(fftSize, dk)  # Units: microns

        opdVals = opdVals[:indexOpdStop]

        opd = opd.astype(self.data.dtype)  # Make sure to upscale precision
        opdVals = opdVals.astype(self.data.dtype)
        return opd, opdVals

    def getRMSFromOPD(self, lowerOPD: float, upperOPD: float, useHannWindow: bool = False) -> np.ndarray:
        """
        Use Parseval's Theorem to calculate our signal RMS from the OPD (magnitude of fourier transform). This allows us to calculate RMS using only contributions
        from certain OPD ranges which ideally are correlated with a specific depth into the sample. In practice the large frequency leakage due to our limited
        bandwidth of measurement causes this assumption to break down, but it can still be useful if taken with a grain of salt.

        Args:
            lowerOPD: RMS will be integrated starting at this lower limit of OPD. Note for a reflectance setup like PWS `sampleDepth = OPD / (2 * meanSampleRI)`
            upperOPD: RMS will be integrated up to this upper OPD limit.
            useHannWindow: If False then use no windowing on the FFT to calculate OPD. If True then use and Hann window.

        Returns:
            A 2d numpy array of the signal RMS at each XY location in the image.
        """
        opd, opdIndex = self.getOpd(useHannWindow=useHannWindow)
        startOpdIdx = np.argmin(np.abs(opdIndex - lowerOPD))  # The index associated with lowerOPD
        stopOpdIdx = np.argmin(np.abs(opdIndex - upperOPD))  # The index associated with upperOPD
        opdSquaredSum = np.sum(opd[:, :, startOpdIdx:stopOpdIdx+1] ** 2, axis=2)  # Parseval's theorem tells us that this is equivalent to the sum of the squares of our original signal
        opdSquaredSum *= len(self.wavenumbers) / opd.shape[2]  # If the original data and opd were of the same length then the above line would be correct. Since the fft may have been upsampled. we need to normalize.
        return np.sqrt(opdSquaredSum)

    @staticmethod
    def fromOpd(opd: np.ndarray, xVals: np.ndarray, useHannWindow: bool):
        """WARNING: This function is untested. it almost certainly doesn't work. Create a KCube from and opd in the form returned by KCube.getOpd. This is useful if you want to do spectral manipulation and then transform back."""
        assert len(xVals.shape) == 1
        fftSize = int(2 ** (np.ceil(np.log2((2 * len(xVals)) - 1))))  # %This is the next size of fft that is  at least 2x greater than is needed but is a power of two. Results in interpolation, helps amplitude accuracy and fft efficiency.
        if useHannWindow: w = np.hanning(len(xVals))
        else: w = np.ones((len(xVals)))
        sig = np.fft.irfft(opd * w[None, None, :], n=fftSize, axis=2)
        #I don't think we need to normalize by the number of elements like we do in getOpd

        # by multiplying by Hann window we reduce the total power of signal. To account for that,
        sig = np.abs(sig / np.sqrt(np.mean(w ** 2)))

        maxWavenumberInterval = 2*np.pi / (xVals[1] - xVals[0])
        dWavenumber = maxWavenumberInterval / len(xVals)
        waveNumbers = len(xVals) / 2 * np.array(range(fftSize // 2 + 1)) * dWavenumber / (fftSize // 2 + 1)
        return sig, waveNumbers


    def getAutoCorrelation(self, isAutocorrMinSub: bool, stopIndex: int) -> t_.Tuple[np.ndarray, np.ndarray]:
        """The autocorrelation of a signal is the covariance of a signal with a
        lagged version of itself, normalized so that the covariance at
        zero-lag is equal to 1.0 (c[0] = 1.0).  The same process without
        normalization is the autocovariance.

        A fast method for determining the autocovariance of a signal with
        itself is to utilize fast-fourier transforms.  In this method, the
        signal is converted to the frequency domain using fft.  The
        frequency-domain signal is then convolved with itself.  The inverse
        fft is performed on this self-convolution, yielding the
        autocorrelation.

        In this instance, the autocorrelation is determined for a series of
        lags, Z. Z is equal to [-P+1:P-1], where P is the quantity of
        measurements in each signal (the quantity of wavenumbers).  Thus, the
        quantity of lags is equal to (2*P)-1.  The fft process is fastest
        when performed on signals with a length equal to a power of 2.  To
        take advantage of this property, a Z-point fft is performed on the
        signal, where Z is a number greater than (2*P)-1 that is also a power
        of 2."""
        fftSize = int(2 ** (np.ceil(np.log2((2 * len(
            self.wavenumbers)) - 1))))  # This is the next size of fft that is  at least 2x greater than is needed but is a power of two. Results in interpolation, helps amplitude accuracy and fft efficiency.

        # Determine the fft for each signal.  The length of each signal's fft
        # will be fftSize.
        cubeFft = np.fft.rfft(self.data, n=fftSize, axis=2)

        # Determine the ifft of the cubeFft.  The resulting ifft of each signal
        # will be of length fftSize..
        cubeAutocorr = np.fft.irfft(np.abs(cubeFft) ** 2, axis=2)  # This is the autocovariance.
        # Obtain only the lags desired.
        # Then, normalize each autocovariance so the value at zero-lags is 1.
        cubeAutocorr = cubeAutocorr[:, :, :len(self.wavenumbers)]
        cubeAutocorr /= cubeAutocorr[:, :, 0, np.newaxis]

        # In some instances, minimum subtraction is desired.  In this case,
        # determine the minimum of each signal and subtract that value from
        # each value in the signal.
        if isAutocorrMinSub:
            cubeAutocorr -= cubeAutocorr.min()

        # Convert the lags from units of indices to wavenumbers.
        lags = np.array(self.wavenumbers) - min(self.wavenumbers)

        # Square the lags. This is how it is in the paper. I'm not sure why though.
        lagsSquared = lags ** 2

        # Before taking the log of the autocorrelation, zero values must be
        # modified to prevent outputs of "inf" or "-inf".
        cubeAutocorr[cubeAutocorr == 0] = 1e-323

        # Obtain the log of the autocorrelation.
        cubeAutocorrLog = np.log(cubeAutocorr)

        # A first-order polynomial fit is determined between lagsSquared and
        # and cubeAutocorrLog.  This fit is to be performed only on the first
        # linear-portion of the lagsSquared vs. cubeAutocorrLog relationship.
        # The index of the last point to be used is indicated by stopIndex.
        lagsSquared = lagsSquared[:stopIndex]
        cubeAutocorrLog = cubeAutocorrLog[:, :, :stopIndex]
        cubeAutocorrLog = np.moveaxis(cubeAutocorrLog, 2, 0)
        cubeAutocorrLog = cubeAutocorrLog.reshape(
            (cubeAutocorrLog.shape[0], cubeAutocorrLog.shape[1] * cubeAutocorrLog.shape[2]))

        # Determine the first-order polynomial fit for each cubeAutocorrLag.
        V = np.stack([np.ones(lagsSquared.shape), lagsSquared])
        V = V.T
        M = np.matmul(V, np.linalg.pinv(V))
        cubeLinear = np.matmul(M, cubeAutocorrLog)
        cubeSlope = (cubeLinear[1, :] - cubeLinear[0, :]) / (lagsSquared[1] - lagsSquared[0])
        cubeSlope = cubeSlope.reshape(self.data.shape[0], self.data.shape[1])
        # -- Coefficient of Determination
        # Obtain the mean of the observed data
        meanObserved = cubeAutocorrLog.mean(axis=0)
        # Obtain the regression sum of squares.
        ssReg = ((cubeLinear - meanObserved) ** 2).sum(axis=0)
        # Obtain the residual sum of squares.
        ssErr = ((cubeAutocorrLog - cubeLinear) ** 2).sum(axis=0)
        # Obtain the total sume of squares.
        ssTot = ssReg + ssErr
        # Obtain rSquared.
        rSquared = ssReg / ssTot
        rSquared = rSquared.reshape(self.data.shape[0], self.data.shape[1])

        cubeSlope = cubeSlope.astype(self.data.dtype)#Make sure to to upscale precision
        rSquared = rSquared.astype(self.data.dtype)
        return cubeSlope, rSquared

    @classmethod
    def fromHdfDataset(cls, dataset: h5py.Dataset):
        """
        Load the KCube object from an `h5py.Dataset` in an HDF5 file

        Args:
            dataset: The `h5py.Dataset` that the KCube data is stored in.
        Returns:
            KCube: A new instance of this class."""
        arr, index = cls.decodeHdf(dataset)
        return cls(arr, index)

    def __add__(self, other):
        ret = self._add(other)
        return KCube(ret, self.wavenumbers, metadata=self.metadata)

    def __sub__(self, other):
        ret = self._sub(other)
        return KCube(ret, self.wavenumbers, metadata=self.metadata)

    def __mul__(self, other):
        ret = self._mul(other)
        return KCube(ret, self.wavenumbers, metadata=self.metadata)

    def __truediv__(self, other):
        ret = self._truediv(other)
        return KCube(ret, self.wavenumbers, metadata=self.metadata)


class FluorescenceImage:
    """
    Represents a fluorescence image taken by the PWS acquisition software.

    Args:
        data: A 2D array of image data.
        md: The metadata object associated with this image.
    """
    def __init__(self, data: np.ndarray, md: pwsdtmd.FluorMetaData):
        self.data = data
        self.metadata = md

    @classmethod
    def fromTiff(cls, directory: str, acquisitionDirectory: t_.Optional[pwsdtmd.Acquisition] = None) -> FluorescenceImage:
        """
        Load an image from a TIFF file.

        Args:
            directory: The path to the folder containing the TIFF file.
            acquisitionDirectory: The `Acquisition` object associated with this acquisition.

        Returns:
            A new instanse of `FluorescenceImage`.
        """
        md = pwsdtmd.FluorMetaData.fromTiff(directory, acquisitionDirectory) #This will raise an error if the folder isn't valid
        return cls.fromMetadata(md)

    @classmethod
    def fromMetadata(cls, md: pwsdtmd.FluorMetaData, lock: t_.Optional[mp.Lock] = None) -> FluorescenceImage:
        """
        Load an image from the metadata object.

        Args:
            md: The metadata object to load the image from.
            lock: An optional multiprocessing `Lock` object to synchronize file access in multiprocessing contexts

        Returns:
            A new instance of `FluorescenceImage`.
        """
        path = os.path.join(md.filePath, pwsdtmd.FluorMetaData.FILENAME)
        if lock is not None:
            lock.acquire()
        try:
            img = tf.TiffFile(path)
        finally:
            if lock is not None:
                lock.release()
        return cls(img.asarray(), md)

    def toTiff(self, directory: str):
        """
        Save this object to a TIFF file.

        Args:
            directory: The path to the folder to save the new file to.
        """
        with open(os.path.join(directory, pwsdtmd.FluorMetaData.FILENAME), 'wb') as f:
            tf.imsave(f, self.data)
        with open(os.path.join(directory, pwsdtmd.FluorMetaData.MDPATH), 'w') as f:
            json.dump(self.metadata, f)

class _FFTHelper:
    class Normalization(Enum):
        POWER = 1
        AMPLITUDE = 2

    @staticmethod
    def getFFTMagnitude(data: np.ndarray, useHannWindow: bool = False, normalization: Normalization = Normalization.POWER):
        """
        Apply windowing, calculate FFT and normalize FFT for the last axis of a real-valued numpy array.

        Args:
            data: A numpy array. The FFT will be calculated along the last axis of the array. The values of this array must be real.
            useHannWindow: If True then a Hann window will be applied to the data before the FFT and the proper normalization will be applied.
            normalization:  When windowing is used the FFT must be normalized using one of two normalizations. "Amplitude" normalization will maintain
                the peak height of detected frequencies but will not preserve the total area under the curve which is asssociated with the energy/power of the signal.
                "Power" normalization will preserve the total area under the curve (important when calculating RMS from an OPD signal) but the amplitudes of detected
                frequencies will generally decrease due to the widened bandwidth associated with windowing.

        Returns:
            A numpy array with the same number of dimensions as `data`. The last axis of the array will be the magnitude of the FFT of the along the last
                axis of the `data` array. Note that the length of the last axis will be longer than the last axis of the `data` array due to FFT interpolation.
        """
        dataLength = data.shape[-1]
        fftSize = int(2 ** (np.ceil(np.log2((2 * dataLength) - 1))))  # This is the next size of fft that is  at least 2x greater than is needed but is a power of two. Results in interpolation, helps amplitude accuracy and fft efficiency.
        fftSize *= 2  # We double the fftsize for even more iterpolation. Not sure why, but that's how it was done in the original matlab code.
        if useHannWindow:  # if hann window checkbox is selected, create hann window
            w = np.hanning(dataLength)  # Hanning window
        else:
            w = np.ones((dataLength))  # Create unity window

        # Calculate the Fourier Transform of the signal multiplied by Hann window
        fft = np.fft.rfft(data * w, n=fftSize, axis=data.ndim-1)
        fft = np.abs(fft)  # We're only interested in the magnitude.
        # Normalize the FFT by the quantity of wavelengths.
        fft /= dataLength

        # by multiplying by Hann window we reduce the total power and amplitude of the signal. To account for that,
        if normalization is _FFTHelper.Normalization.POWER:
            fft *= np.sqrt(len(w) / np.sum(w ** 2)) # Correct the signal so the true power (area under curve) is preserved. This will prevent windowing from affecting integration of RMS but the amplitude of each frequency will be reduced https://dsp.stackexchange.com/questions/47598/does-windowing-affect-parsevals-theorem
        elif normalization is _FFTHelper.Normalization.AMPLITUDE:
            fft *= len(w) / np.sum(w)  # Correct amplitude scaling error caused by windowing. Does not preserve energy of signal though. https://www.mathworks.com/matlabcentral/answers/372516-calculate-windowing-correction-factor
        else:
            raise ValueError(f"{normalization} is not a valid normalization.")
        return fft



from __future__ import annotations
import abc
import enum
import json
import typing
from ._treeItem import TreeItem
from pwspy.utility.micromanager import PositionList
import os
from .sequencerCoordinate import SequencerCoordinateRange

StepTypeNames = dict(  # The names to represent the `steps` that can be in an acquisition sequence
    ACQ="Acquisition",
    POS="Multiple Positions",
    TIME="Time Series",
    CONFIG="Configuration Group",
    SUBFOLDER="Enter Subfolder",
    EVERYN="Once per `N` iterations",
    PFS="Optical Focus Lock",
    PAUSE="Pause",
    ROOT="Initialization",
    AF="Software Autofocus",
    ZSTACK="Z-Stack",
    AUTOSHUTTER='AutoShutter',
    WAIT="Wait"
)


class SequencerStep(TreeItem):
    """Implementation of a TreeItem for representing a sequencer step.

    Args:
        id: The unique integer assigned to this step by the acquisition software
        settings: The settings for this step. Saved as JSON in the sequence file.
        stepType: Indicates what type of step this is. E.g. acquisition, time-series, focus lock, etc.
        children: A list of steps which are direct children of this step.
    """
    def __init__(self, id: int, settings: dict, stepType: str, children: typing.List[SequencerStep] = None):
        super().__init__()
        self.id = id
        self.settings = settings
        self._stepType = stepType
        if children is not None:
            self.addChildren(children)

    @property
    def stepType(self) -> str:
        return self._stepType

    @staticmethod
    def hook(dct: dict):
        """
        This method defines how the JSON library should translate from JSON to one of these objects.
        Args:
            dct: The `dict` representing the raw representation of the JSON\
        """
        if all([i in dct for i in ("id", 'stepType', 'settings')]):
            clazz = _SequencerStepTypes[dct['stepType']].value
            s = clazz(**dct)
            return s
        else:
            return dct

    @staticmethod
    def fromJson(j: str) -> SequencerStep:
        return json.loads(j, object_hook=SequencerStep.hook)
    
    @staticmethod
    def fromJsonFile(filePath: str) -> SequencerStep:
        with open(filePath, 'r') as f:
            return json.loads(f.read(), object_hook=SequencerStep.hook)
    
    def __repr__(self):  # Text representation of this object.
        return str(f"Step {self.id}: {self.stepType}")

    def printSubTree(self, _indent: int = 0) -> None:
        """
        Print out this step and all sub-steps in a human-readable format.
        """
        indent = ''.join((['\t'] * _indent))
        print(f"{indent}{self}")
        for child in self.children():
            child.printSubTree(_indent=_indent+1)

    def getCoordinate(self) -> SequencerCoordinateRange:
        """Returns a sequencer coordinate range that points to this steps location in the tree of steps."""
        return SequencerCoordinateRange([(step.id, None) for step in self.getTreePath()])

    def __getitem__(self, item: int):
        """Enable [i] subscripting to get a child step"""
        return self.children()[item]


class ContainerStep(SequencerStep):
    """
    A class for steps which can contain other steps within it.
    """
    pass


class IterableSequencerStep(ContainerStep):
    """
    A base-class for steps which are iterable. Despite only being a single step they run multiple times in an acquisition.
    This add some complications as we want to keep track of which iteration the sub-steps of this belong to.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @abc.abstractmethod
    def stepIterations(self):
        """
        Return the total number of iterations of this step.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def getIterationName(self, iteration: int) -> str:
        """
        Return the name associated with `iteration` E.G. for a multiple-positions step this will be the name assigned to the position in the position list.
        Args:
            iteration: The iteeration number we are interested in.

        Returns: A name for the requested iteration.

        """
        raise NotImplementedError()


class PositionsStep(IterableSequencerStep):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._positionList = PositionList.fromDict(self.settings['posList'])

    def stepIterations(self):
        if not hasattr(self, '_len'):
            self._len = len(self._positionList)
        return self._len

    def getIterationName(self, iteration: int) -> str:
        return self._positionList[iteration].label

    def getPositionList(self) -> PositionList:
        return self._positionList


class TimeStep(IterableSequencerStep):
    def stepIterations(self):
        if not hasattr(self, '_len'):
            self._len = self.settings['numFrames']
        return self._len

    def getIterationName(self, iteration: int) -> str:
        return f"{iteration * self.settings['frameIntervalMinutes']} min."


class ZStackStep(IterableSequencerStep):
    def stepIterations(self):
        if not hasattr(self, '_len'):
            self._len = self.settings['numStacks']
        return self._len

    def getIterationName(self, iteration: int) -> str:
        return f"{iteration * self.settings['intervalUm']} μm"


class _SequencerStepTypes(enum.Enum):
    """
    An enumerator containing the sub-class of `SequencerStep` to use for each type of step.
    """
    ACQ = SequencerStep
    PFS = ContainerStep
    POS = PositionsStep
    TIME = TimeStep
    AF = SequencerStep
    CONFIG = ContainerStep
    PAUSE = SequencerStep
    EVERYN = ContainerStep
    ROOT = ContainerStep
    SUBFOLDER = ContainerStep
    ZSTACK = ZStackStep
    AUTOSHUTTER = ContainerStep
    WAIT = SequencerStep


class RuntimeSequenceSettings:
    """
    This represents the object saved when a new acquisition is run.
    """
    FILENAME = "sequence.rtpwsseq"
    OLDFILENAME = "sequence.pwsseq"

    def __init__(self, uuid: str, dateString: str, rootStep: SequencerStep):
        self.uuid = uuid
        self.dateString = dateString
        self.rootStep = rootStep

    @classmethod
    def fromJsonFile(cls, directory: str):
        if os.path.exists(os.path.join(directory, cls.FILENAME)):
            with open(os.path.join(directory, cls.FILENAME), 'r') as f:
                return cls(**json.load(f, object_hook=SequencerStep.hook))
        elif os.path.exists(os.path.join(directory, cls.OLDFILENAME)):
            with open(os.path.join(directory, cls.OLDFILENAME)) as f:
                rootStep = json.load(f, object_hook=SequencerStep.hook)
            return cls(uuid=None, dateString=None, rootStep=rootStep)  # Old sequence files didn't have a uuid or date. Too bad
        else:
            raise FileNotFoundError(f"No valid sequence file found in: {directory}")


# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

# -*- coding: utf-8 -*-
"""
Created on Mon Dec  3 17:53:24 2018

@author: Nick Anthony
"""
from __future__ import annotations

import abc
import json
import typing
import warnings
from dataclasses import dataclass
from numbers import Number
from typing import Union
import numpy as np
import copy
import matplotlib.pyplot as plt
import matplotlib as mpl
import scipy.io as spio
from pwspy.utility.micromanager.PropertyMap import PropertyMap, PropertyMapArray, Property, PropertyArray

@dataclass
class Position1d:
    """A 1D position usually describing the position of a Z-axis translation stage.

    Attributes:
        z: The position
        zStage: Then name of the translation stage
    """
    z: float    
    stageName: str = ''
    numAxes: int = 1

    def __post_init__(self):
        assert isinstance(self.z, float)
        assert isinstance(self.stageName, str)

    @staticmethod
    def fromDict(d: dict) -> Position1d:
        return Position1d(**d)

    def toPropertyMap(self) -> PropertyMap:
        return PropertyMap({"Device": Property(self.stageName),
                            "Position_um": PropertyArray([Property(self.z)])})

    @staticmethod
    def fromPropertyMap(pmap: PropertyMap) -> Position1d:
        if len(pmap['Position_um']) != 1:
            raise Exception("RERAR")
        return Position1d(z=pmap['Position_um'][0].value, stageName=pmap['Device'].value)

    def __repr__(self):
        return f"Position1d({self.stageName}, {self.z})"


@dataclass
class Position2d:
    """Represents a 2D position for a single xy stage in micromanager.

    Attributes:
        x: The x position
        y: The y position
        xyStage: The name of the 2 dimensional translation stage
    """
    x: float
    y: float
    stageName: str = ''
    numAxes: int = 2

    def __post_init__(self):
        assert isinstance(self.x, Number)
        assert isinstance(self.y, Number)
        assert isinstance(self.stageName, str)
        self.x = float(self.x)
        self.y = float(self.y)  # make sure to convert in case it is a numpy type

    @staticmethod
    def fromDict(d: dict) -> Position2d:
        return Position2d(**d)

    def toPropertyMap(self) -> PropertyMap:
        return PropertyMap({"Device": Property(self.stageName),
                            "Position_um": PropertyArray([Property(self.x), Property(self.y)])})

    @staticmethod
    def fromPropertyMap(pmap: PropertyMap) -> Position2d:
        if len(pmap['Position_um']) != 2:
            raise Exception("Errr")
        x, y = pmap['Position_um'][0].value, pmap['Position_um'][1].value
        return Position2d(x=x, y=y, stageName=pmap['Device'].value)

    def mirrorX(self) -> Position2d:
        self.x *= -1
        return self

    def mirrorY(self) -> Position2d:
        self.y *= -1
        return self

    def renameStage(self, newName):
        self.stageName = newName

    def __repr__(self):
        return f"Position2d({self.stageName}, {self.x}, {self.y})"

    def __add__(self, other: Union[PositionList, Position2d, MultiStagePosition]) -> Union[PositionList, Position2d, MultiStagePosition]:
        if isinstance(other, PositionList):
            return other.__add__(self)
        elif isinstance(other, MultiStagePosition):
            return other.__add__(self)
        elif isinstance(other, Position2d):
            return Position2d(self.x + other.x,
                              self.y + other.y,
                              self.stageName)
        else:
            raise TypeError(f"Type {type(other)} is not supported.")

    def __sub__(self, other: Union[PositionList, Position2d, MultiStagePosition]) -> Union[PositionList, Position2d, MultiStagePosition]:
        if isinstance(other, PositionList):
            return other.copy().mirrorX().mirrorY().__add__(self)  # a-b == -b + a
        elif isinstance(other, MultiStagePosition):
            other = other.copy()  # Don't change the original object
            other.getXYPosition().mirrorX().mirrorY()  # invert the object.
            return other.__add__(self)  # a-b == -b + a
        elif isinstance(other, Position2d):
            return Position2d(self.x - other.x,
                              self.y - other.y,
                              self.stageName)
        else:
            raise TypeError(f"Type {type(other)} is not supported.")

    def __eq__(self, other: 'Position2d'):
        return all([self.x == other.x,
                    self.y == other.y,
                    self.stageName == other.stageName])


@dataclass
class MultiStagePosition:
    """Mirrors the class of the same name from Micro-Manager. Can contain multiple Positon1d or Position2d objects.
    Ideal for a system with multiple translation stages. It is assumed that there is only a single 2D stage and a single
    1D stage.

    Attributes:
        label: A name for the position
        defaultXYStage: The name of the 2D stage
        defaultZStage: The name of the 1D stage
        stagePositions: A list of `Position1d` and `Position2D` objects, usually just one of each.
    """
    label: str
    defaultXYStage: str
    defaultZStage: str
    stagePositions: typing.List[typing.Union[Position1d, Position2d]]
    gridRow: int = 0
    gridCol: int = 0

    @staticmethod
    def fromDict(d: dict) -> MultiStagePosition:
        sps = []
        for i in d['stagePositions']:
            if i['numAxes'] == 1:
                sps.append(Position1d.fromDict(i))
            elif i['numAxes'] == 2:
                sps.append(Position2d.fromDict(i))
            else:
                raise Exception()
        return MultiStagePosition(label=d['label'],
                                  defaultXYStage=d['defaultXYStage'],
                                  defaultZStage=d['defaultZStage'],
                                  stagePositions=sps,
                                  gridRow=d['gridRow'],
                                  gridCol=d['gridCol'])

    def toPropertyMap(self) -> PropertyMap:
        return PropertyMap({
            "DefaultXYStage": Property(self.defaultXYStage),
            "DefaultZStage": Property(self.defaultZStage),
            "DevicePositions": PropertyMapArray([i.toPropertyMap() for i in self.stagePositions]),
            "GridCol": Property(0),
            "GridRow": Property(0),
            "Label": Property(self.label),
            "Properties": PropertyMap({})
        })

    @staticmethod
    def fromPropertyMap(d: PropertyMap) -> MultiStagePosition:
        positions = []
        for i in d['DevicePositions']:
            if len(i["Position_um"]) == 1:
                positions.append(Position1d.fromPropertyMap(i))
            elif len(i["Position_um"]) == 2:
                positions.append(Position2d.fromPropertyMap(i))
            else:
                raise Exception("EEEEE")
        return MultiStagePosition(label=d['Label'].value, defaultXYStage=d['DefaultXYStage'].value, defaultZStage=d['DefaultZStage'].value,
                                  stagePositions=positions)

    def getXYPosition(self):
        """Return the first `Position2d` saved in the `positions` list"""
        d1pos = [i for i in self.stagePositions if isinstance(i, Position2d)]
        return [i for i in d1pos if i.stageName == self.defaultXYStage][0]

    def getZPosition(self):
        """Return the first `Position1d` saved in the positions` list. Returns `None` if no position is found."""
        try:
            d1pos = [i for i in self.stagePositions if isinstance(i, Position1d)]
            return [i for i in d1pos if i.stageName == self.defaultZStage][0]
        except IndexError:
            return None

    def renameXYStage(self, label: str):
        """Change the name of the xy stage.

        Args:
            label: The new name for the xy Stage
        """
        self.getXYPosition().renameStage(label)
        self.defaultXYStage = label  # This must come after the previous line or things get messed up

    def copy(self) -> MultiStagePosition:
        """Creates a copy fo the object

        Returns:
            A new `MultiStagePosition` object.
        """
        return copy.deepcopy(self)

    def __add__(self, other: Union[Position2d, MultiStagePosition, PositionList]) -> Union[MultiStagePosition, PositionList]:
        """Allow adding a position to another position or `PositionList`. Doesn't work on the Z-Axis but will sum together
        the X and Y axis values. Used to translate a position by an offset.

        Args:
            other: The object whose XY coordinates should be added to this objects XY coordinates.
        """
        if isinstance(other, Position2d):
            newPos = self.getXYPosition().__add__(other)
            positions = copy.copy(self.stagePositions)
            positions.remove(self.getXYPosition())
            positions.append(newPos)
            return MultiStagePosition(self.label, self.defaultXYStage, self.defaultZStage, stagePositions=positions)
        elif isinstance(other, MultiStagePosition):
            return self.__add__(other.getXYPosition())
        elif isinstance(other, PositionList):
            return other.__add__(self)
        else:
            raise NotImplementedError

    def __sub__(self, other: Union[Position2d, MultiStagePosition, PositionList]) -> Union[MultiStagePosition, PositionList]:
        """See the documentation for __add__"""
        if isinstance(other, Position2d):
            newPos = self.getXYPosition().__sub__(other)
            positions = copy.copy(self.stagePositions)
            positions.remove(self.getXYPosition())
            positions.append(newPos)
            return MultiStagePosition(self.label, self.defaultXYStage, self.defaultZStage, stagePositions=positions)
        elif isinstance(other, MultiStagePosition):
            return self.__sub__(other.getXYPosition())
        elif isinstance(other, PositionList):
            return other.copy().mirrorX().mirrorY().__add__(self)  # a-b == -b + a
        else:
            raise NotImplementedError

    def __eq__(self, other: MultiStagePosition):
        """Returns True if the stage names and stage coordinates are equivalent."""
        return all([self.defaultXYStage == other.defaultXYStage,
                    self.defaultZStage == other.defaultZStage,
                    self.getXYPosition() == other.getXYPosition(),
                    self.getZPosition() == other.getZPosition()])

    def __repr__(self):
        s = f"MultiStagePosition({self.label}, "
        for i in self.stagePositions:
            s += '\n\t' + i.__repr__()
        return s


class PositionList:
    """Represents a micromanager positionList. can be loaded from and saved to a micromanager .pos file.

    Args:
        positions: A list of `MultiStagePosition` objects

    Attributes:
        positions: A list of `MultiStagePosition` objects
    """

    def __init__(self, positions: typing.List[MultiStagePosition]):
        super().__init__()
        assert isinstance(positions, list)
        assert isinstance(positions[0], MultiStagePosition)
        self.positions = positions

    @staticmethod
    def fromDict(d: dict):
        p = []
        for i in d['positions']:
            p.append(MultiStagePosition.fromDict(i))
        return PositionList(p)

    def mirrorX(self) -> PositionList:
        """Invert all x coordinates

        Returns:
            A reference to this object.
        """
        for i in self.positions:
            i.getXYPosition().mirrorX()
        return self

    def mirrorY(self) -> PositionList:
        """Invert all y coordinates

         Returns:
             A reference to this object.
         """
        for i in self.positions:
            i.getXYPosition().mirrorY()
        return self

    def renameStage(self, label) -> PositionList:
        """Change the name of the xy stage.

        Args:
            label: The new name for the xy Stage

        Returns:
            A reference to this object
        """
        for i in self.positions:
            i.renameXYStage(label)
        return self

    def copy(self) -> PositionList:
        return copy.deepcopy(self)

    @staticmethod
    def fromPropertyMap(pmap: PropertyMap) -> PositionList:
        """Attempt to load a PositionList from a PropertyMap. May throw an exception."""
        if isinstance(pmap, PropertyMap):
            if "StagePositions" in pmap:
                return PositionList([MultiStagePosition.fromPropertyMap(i) for i in pmap["StagePositions"]])
        raise Exception("JsonParseException")

    def toPropertyMap(self) -> PropertyMap:
        """Returns the position list as a PropertyMap that is formatted just like a `PropertyMap` from Micro-Manager."""
        pmap = PropertyMap({"StagePositions": PropertyMapArray([i.toPropertyMap() for i in self.positions])})
        return pmap

    @classmethod
    def fromNanoMatFile(cls, path: str, xyStageName: str):
        """Load an instance of the `PositionList` from a file saved by NC MATLAB acquisition software.

        Args:
            path: The file path to the .mat file.
            xyStageName: To adapt the MATLAB file format to the Micro-Manager we need to manually supply a name for the
            XY stage

        Returns:
            A new instance of `PositionList`
        """
        mat = spio.loadmat(path)
        l = mat['list']
        positions = []
        for i in range(l.shape[0]):
            coordString: str = l[i][0][0]
            x, y = coordString[1:-1].split(',')
            x, y = float(x), float(y)
            pos = Position2d(x, y, xyStageName)
            positions.append(MultiStagePosition(str(i), xyStageName, '', [pos]))
        return PositionList(positions)

    def toNanoMatFile(self, path: str):
        """Save this object to a .mat file in the format saved by NC MATLAB acquistion software.

        Args:
            path: The file path for the new .mat file.
        """
        matPositions = []
        for pos in self.positions:
            pos = pos.getXYPosition()
            matPositions.append(f"({pos.x}, {pos.y})")
        matPositions = np.asarray(matPositions, dtype=np.object)
        print(matPositions.shape)
        spio.savemat(path, {'list': matPositions[:, None]})

    def getAffineTransform(self, otherList: PositionList) -> np.ndarray:
        """
        Calculate the partial affine transformation between this position list and another position list. Both position lists must have the same length

        Args:
            otherList (PositionList): A position list of the same length as this position list. Each position is assumed to correspond to the position of the
                same index in this list.

        Returns:
            np.ndarray: A 2x3 array representing the partial affine transform (rotation, scaling, and translation, but no skew)

        Examples:
            a = PositionList.fromNanoMatFile(r'F:/Data/AirDryingSystemComparison/NanoPreDry/corners/positions.mat', "TIXYDRIVE")
            b = PositionList.fromNanoMatFile(r'F:/Data/AirDryingSystemComparison/NanoPostDry/corners/positions.mat',"TIXYDRIVE")
            t = a.getAffineTransform(b)
            origPos = PositionList.fromNanoMatFile(r'F:/Data/AirDryingSystemComparison/NanoPreDry/0_8NA/position_list1.mat',"TIXYDRIVE")
            newPos = origPos.applyAffineTransform(t)
        """
        import cv2
        assert len(otherList) == len(self)
        selfXY = [pos.getXYPosition() for pos in self.positions]
        otherXY = [pos.getXYPosition() for pos in otherList.positions]
        selfArr = np.array([(pos.x, pos.y) for pos in selfXY], dtype=np.float32)
        otherArr = np.array([(pos.x, pos.y) for pos in otherXY], dtype=np.float32)
        transform, inliers = cv2.estimateAffine2D(selfArr, otherArr)
        if inliers.sum() < inliers.shape[0]:
            warnings.warn(f"Only {inliers.sum()} of {inliers.shape[0]} coordinates were considered to be inliers.")
        return transform

    def applyAffineTransform(self, t: np.ndarray):
        """Given an affine transformation array this method will transform all positions in this position list.
        Args:
            t (np.ndarray): A 2x3 array representing the partial affine transform (rotation, scaling, and translation, but no skew)
        """

        import cv2
        assert isinstance(t, np.ndarray)
        assert t.shape == (2, 3)
        selfXY = [pos.getXYPosition() for pos in self.positions]
        selfArr = np.array([[pos.x, pos.y] for pos in selfXY], dtype=np.float32)
        selfArr = selfArr[None, :, :]  # This is needed for opencv transform to work
        ret = cv2.transform(selfArr, t)
        positions = []
        for i in range(len(self)):
            pos = copy.deepcopy(self[i].getXYPosition())
            pos.x = float(ret[0, i, 0])  # Convert to float because the numpy number types cause some errors during jsonification.
            pos.y = float(ret[0, i, 1])
            positions.append(MultiStagePosition(self[i].label, self[i].defaultXYStage, '', [pos]))
        return PositionList(positions)

    def __repr__(self):
        s = "PositionList(\n["
        for i in self.positions:
            s += str(i) + '\n'
        s += '])'
        return s

    def __add__(self, other: Union[Position2d, MultiStagePosition, PositionList]) -> PositionList:
        if isinstance(other, Position2d):
            return PositionList([i + other for i in self.positions])
        elif isinstance(other, MultiStagePosition):
            self.__add__(other.getXYPosition())
        else:
            assert len(other) == len(self), "Cannot subtract position lists of different sizes"
            return PositionList([a + b for a, b in zip(self.positions, other.positions)])

    def __sub__(self, other: Union[Position2d, MultiStagePosition, PositionList]) -> PositionList:
        if isinstance(other, Position2d):
            return PositionList([i - other for i in self.positions])
        elif isinstance(other, MultiStagePosition):
            return self.__sub__(other.getXYPosition())
        else:
            assert len(other) == len(self), "Cannot subtract position lists of different sizes"
            return PositionList([a - b for a, b in zip(self.positions, other.positions)])

    def __len__(self):
        return len(self.positions)

    def __getitem__(self, idx: Union[slice, int]) -> MultiStagePosition:
        return self.positions[idx]

    def __eq__(self, other: PositionList):
        return all([len(self) == len(other)] +
                   [self[i] == other[i] for i in range(len(self))])

    def plot(self, fig: plt.Figure, ax: plt.Axes):
        """Open a matplotlib plot showing the positions contained in this list."""
        annot = ax.annotate("", xy=(0, 0), xytext=(20, 20), textcoords="offset points",
                            bbox=dict(boxstyle="round", fc="w"),
                            arrowprops=dict(arrowstyle="->"))
        annot.set_visible(False)
        ax.set_xlabel("x")
        ax.set_ylabel('y')
        ax.set_aspect('equal')
        cmap = mpl.cm.get_cmap("gist_rainbow")
        colors = [cmap(i) for i in np.linspace(0, 1, num=len(self.positions))]
        names = [pos.label for pos in self.positions]
        sc = ax.scatter([pos.getXYPosition().x for pos in self.positions], [pos.getXYPosition().y for pos in self.positions],
                         c=[colors[i] for i in range(len(self.positions))])

        def update_annot(ind):
            pos = sc.get_offsets()[ind["ind"][0]]
            annot.xy = pos
            text = "{}, {}".format(" ".join(list(map(str, ind["ind"]))),
                                   " ".join([names[n] for n in ind["ind"]]))
            annot.set_text(text)
            #            annot.get_bbox_patch().set_facecolor(cmap(norm(c[ind["ind"][0]])))
            annot.get_bbox_patch().set_alpha(0.4)

        def hover(event):
            vis = annot.get_visible()
            if event.inaxes == ax:
                cont, ind = sc.contains(event)
                if cont:
                    update_annot(ind)
                    annot.set_visible(True)
                    fig.canvas.draw_idle()
                else:
                    if vis:
                        annot.set_visible(False)
                        fig.canvas.draw_idle()

        fig.canvas.mpl_connect("motion_notify_event", hover)


if __name__ == '__main__':
    path1 = r'C:\Users\nicke\Desktop\PositionList.pos'
    path2 = r'C:\Users\nicke\Desktop\PositionList5.pos'
    p = PropertyMap.loadFromFile(path1)
    pp = PositionList.fromPropertyMap(p)
    pp.mirrorX().mirrorY().mirrorX().mirrorY()
    ppp = pp.toPropertyMap()
    ppp.saveToFile(path2)
    with open(path1) as f1, open(path2) as f2:
        assert f1.read() == f2.read()
    a = 1


    def generateList(data: np.ndarray) -> PositionList:
        assert isinstance(data, np.ndarray)
        assert len(data.shape) == 2
        assert data.shape[1] == 2
        positions = []
        for n, i in enumerate(data):
            positions.append(MultiStagePosition(f'Cell{n + 1}', 'TIXYDrive', 'TIZDrive', stagePositions=[Position2d(i[0], i[1], 'TIXYDrive')]))
        plist = PositionList(positions)
        return plist


    def pws1to2(loadPath, newOriginX, newOriginY):
        if isinstance(loadPath, str):
            pws1 = PositionList.fromPropertyMap(PropertyMap.loadFromFile(loadPath))
        elif isinstance(loadPath, PositionList):
            pws1 = loadPath
        else:
            raise TypeError("loadPath must be either `str` or `PositionList`")
        pws2 = pws1.copy()
        pws2.mirrorX()
        pws2.mirrorY()
        pws2Origin = Position2d(newOriginX, newOriginY)
        offset = pws2Origin - pws2.positions[0]
        pws2 = pws2 + offset
        pws2.renameStage("TIXYDrive")
        return pws2


    def pws1toSTORM(loadPath, newOriginX, newOriginY):
        if isinstance(loadPath, str):
            pws1 = PositionList.fromPropertyMap(PropertyMap.loadFromFile(loadPath))
        elif isinstance(loadPath, PositionList):
            pws1 = loadPath
        else:
            raise TypeError("loadPath must be either `str` or `PositionList`")
        pws2 = pws1.copy()
        pws2.mirrorY()
        pws2Origin = Position2d(newOriginX, newOriginY)
        offset = pws2Origin - pws2.positions[0]
        pws2 = pws2 + offset
        pws2.renameStage("TIXYDrive")
        return pws2


    def pws2toSTORM(loadPath, newOriginX, newOriginY):
        if isinstance(loadPath, str):
            pws2 = PositionList.fromPropertyMap(PropertyMap.loadFromFile(loadPath))
        elif isinstance(loadPath, PositionList):
            pws2 = loadPath
        else:
            raise TypeError("loadPath must be either `str` or `PositionList`")
        storm = pws2.copy()
        storm.mirrorX()
        stormOrigin = Position2d(newOriginX, newOriginY)
        offset = stormOrigin - storm.positions[0]
        storm = storm + offset
        return storm


    def STORMtoPws2(loadPath, newOriginX, newOriginY):
        if isinstance(loadPath, str):
            storm = PositionList.fromPropertyMap(PropertyMap.loadFromFile(loadPath))
        elif isinstance(loadPath, PositionList):
            storm = loadPath
        else:
            raise TypeError("loadPath must be either `str` or `PositionList`")
        pws2 = storm.copy()
        pws2.mirrorX()
        pws2Origin = Position2d(newOriginX, newOriginY)
        offset = pws2Origin - pws2.positions[0]
        pws2 = pws2 + offset
        return pws2


    a = 1


# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

"""
Objects useful for dealing with files saved by Micro-Manager. https://micro-manager.org/

Classes
----------
.. autosummary::
   :toctree: generated/

   Image
   Position1d
   Position2d
   PositionList
   Property
   PropertyMap
   MultiStagePosition

"""
__all__ = ['Image', 'PositionList', 'Position2d', 'Position1d', 'MultiStagePosition']
from .images import Image
from .positions import Position1d, Position2d, PositionList, MultiStagePosition
from .PropertyMap import Property, PropertyMap



# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

from pwspy.dataTypes import DynCube
import h5py

outPath = r''
refPath = r''

ref = DynCube.loadAny(refPath)
ref.data[:, :, :] = ref.data.mean(axis=2)[:, :, None] #The reference should be static over time. Take the mean to filter out all noise.
#TODO we should save or report the noise level as well.
with h5py.File("outputRef.h5py", 'w') as f:
    ref.toHdfDataset(f, outPath)



# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

"""This script detects all local maxima in the spectra of a PWS image and then performs various morphological operations on the peak locations in an
 attempt to localize the cell surface. It is very finnick and needs to be adjusted for each individual image. I reccomend the "activeContour" script instead."""

if __name__ == '__main__':
    from pwspy.dataTypes import Acquisition
    from glob import glob
    import os
    import scipy as sp
    import numpy as np
    from mpl_qt_viz.visualizers import PlotNd, MultiPlot
    import matplotlib.pyplot as plt
    import skimage.morphology as morph
    import skimage.measure as meas
    from examples.findOPDSurface.peakDetection.funcs import prune3dIm, Skel

    # rootDir = r'G:\Data\NA_i_vs_NA_c\matchedNAi_largeNAc\cells'
    rootDir = r'G:\Data\NA_i_vs_NA_c\smallNAi_largeNAc\cells'
    anName = 'p0'
    roiName = 'cell'
    sigma = .5  # Blur radius in microns. greater sigma generally leads to greater sensitivity, with more contiguous regions of detected peaks. Also greater noise though.
    opdCutoffLow = 3
    opdCutoffHigh = 20

    files = glob(os.path.join(rootDir, 'Cell*'))
    acqs = [Acquisition(f) for f in files]

    for acq in acqs:
        try:
            an = acq.pws.loadAnalysis(anName)
        except OSError:
            print("No analysis found. skipping")
            continue

        opd, opdIndex = an.reflectance.getOpd(useHannWindow=True)  # Using the Hann window should improve dynamic range and reduce false peaks from frequency leakage.

        # Remove the low OPD signal.
        idxLow = np.argmin(np.abs(opdIndex - opdCutoffLow))
        idxHigh = np.argmin(np.abs(opdIndex - opdCutoffHigh))
        opd = opd[:, :, idxLow:idxHigh]
        opdIndex = opdIndex[idxLow:idxHigh]

        # Blur laterally to denoise. This step is vital
        Sigma = sigma / acq.pws.pixelSizeUm  # convert from microns to pixels
        for i in range(opd.shape[2]):
            opd[:, :, i] = sp.ndimage.filters.gaussian_filter(opd[:, :, i], Sigma, mode='reflect')

        #Try to account for decreased sensitivity at higher opd
        print("Adjusting OPD signal")
        opd = opd * (opdIndex[None, None, :] ** 0.5)

        print("Detect Peaks")
        arr = np.zeros_like(opd, dtype=bool)
        for i in range(opd.shape[0]):
            for j in range(opd.shape[1]):
                peaks, properties = sp.signal.find_peaks(opd[i, j, :])  # TODO vectorize? use kwargs?
                arr[i, j, :][peaks] = True



        if True: # plotPeakDetection
            print("Plotting random peaks")
            fig, ax = plt.subplots()
            plt.xlabel("OPD (um)")
            artists = []
            for i in range(100):
                x, y = np.random.randint(0, opd.shape[1]), np.random.randint(0, opd.shape[0])
                lines = ax.plot(opdIndex, opd[y, x, :])
                vlines = ax.vlines(opdIndex[arr[y, x, :]], ymin=opd[y, x, :].min(), ymax=opd[y, x, :].max())
                artists.append(lines + [vlines])
            mp = MultiPlot(artists, "Peak Detection")
            mp.show()

        #2d regions
        print("Remove regions. 2D")
        arr2 = np.zeros_like(arr)
        for i in range(arr.shape[2]):
            labeled = meas.label(arr[:, :, i])
            properties = meas.regionprops(labeled)
            for prop in properties:
                if prop.area > 30:
                    coords = prop.coords
                    coords = (coords[:, 0], coords[:, 1], i)
                    arr2[coords] = True

        #3d regions
        print("Remove regions. 3D")
        arr3 = np.zeros_like(arr2)
        labeled = meas.label(arr2)
        properties = meas.regionprops(labeled)
        for prop in properties:
            if prop.area > 1e3:
                coords = prop.coords
                coords = (coords[:, 0], coords[:, 1], coords[:,2])
                arr3[coords] = True

        print("Skeletonize. 2D")
        arr4 = np.zeros_like(arr3)  # 3d skeletonizing didn't work well.
        for i in range(arr3.shape[2]):
            disk = morph.disk(3)
            temp = morph.binary_dilation(arr3[:, :, i], disk)
            arr4[:, :, i] = morph.skeletonize(temp)

        print("analyze skeleton 2d")
        arr5 = prune3dIm(arr4, plotEveryN=15)


        #3d regions
        print("Remove regions. 3D")
        arr6 = np.zeros_like(arr5)
        disk = morph.disk(2)
        _ = morph.binary_dilation(arr5, disk[:,:,None])
        labeled = meas.label(_)
        properties = meas.regionprops(labeled)
        for prop in properties:
            if prop.area > 1e4:
                coords = prop.coords
                coords = (coords[:, 0], coords[:, 1], coords[:,2])
                arr6[coords] = True
        ThreeDskel = morph.skeletonize(arr6) #3d Skeletonized. This is just used for a cool plot. 2d skeletonized is more useful.

        Skel3d = Skel(ThreeDskel)

        temp = np.zeros_like(arr6)  # 2d skeletonized
        for i in range(temp.shape[2]):
            temp[:, :, i] = morph.skeletonize(arr6[:, :, i])
        arr6 = temp

        #Estimate an OPD distance for each pixel. This is too slow.
        print("Condense to 2d")
        height = np.zeros((arr6.shape[0], arr6.shape[1]))
        for i in range(arr6.shape[0]):
            for j in range(arr6.shape[1]):
                where = np.argwhere(arr6[i, j, :]).squeeze()
                if where.shape == (0,):
                    height[i, j] = np.nan
                else:
                    height[i, j] = opdIndex[where].max()

        print("Interpolate")
        # Interpolate out the Nans
        def interpolateNans(arr, method='linear'):  # 'cubic' may also be good.
            """Interpolate out nan values along the third axis of an array"""
            from scipy.interpolate import griddata
            x, y = list(range(arr.shape[1])), list(range(arr.shape[0]))
            X, Y = np.meshgrid(x, y)
            nans = np.isnan(arr)
            Xna = X[~nans]
            Yna = Y[~nans]
            Zna = arr[~nans]

            coords = np.array(list(zip(Xna.flatten(), Yna.flatten())))
            icoords = np.array(list(zip(X.flatten(), Y.flatten())))
            Z = griddata(coords, Zna.flatten(), icoords, method=method)
            Z[np.isnan(Z)] = 0
            Z = Z.reshape(arr.shape)

            return Z

        Z = interpolateNans(height, method='linear')

        print("Plot")
        indices = [range(opd.shape[0]), range(opd.shape[1]), opdIndex]
        p = PlotNd(opd, indices=indices, names=('y','x','opd'))
        p2 = PlotNd(arr, indices=indices, names=('y','x','opd'))
        p4 = PlotNd(arr3, indices=indices, names=('y','x','opd'))
        p5 = PlotNd(arr4, indices=indices, names=('y','x','opd'))
        p6 = PlotNd(arr5, indices=indices, names=('y','x','opd'))
        p7 = PlotNd(arr6, indices=indices, names=('y','x','opd'))


        fig = plt.figure()
        plt.imshow(an.meanReflectance, cmap='gray', clim=[np.percentile(an.meanReflectance, 1), np.percentile(an.meanReflectance, 99)])
        fig.show()

        fig = plt.figure()
        _ = height.copy()
        _[np.isnan(_)]=0  # The nans don't plot very well
        plt.imshow(_, cmap='jet', clim=[_.min(), np.percentile(_,99)])
        plt.colorbar()
        fig.show()

        Skel3d.plot()

        fig = plt.figure()
        _ = Z.copy()
        _[_==0] = np.percentile(_[_!=0], .1)
        plt.imshow(_)
        plt.colorbar()
        plt.title("Simple Interpolation.")
        fig.show()


        print("Done")
        a = 1  # debug here
    b = 1 # debug breakpoint here


# -*- coding: utf-8 -*-
# Copyright 2018-2021 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

"""
Load the OPD from a previously saved analysis result and plot it using a special multi-dimensional plotting widget.

@author: Nick Anthony
"""

import pwspy.dataTypes as pwsdt
from mpl_qt_viz.visualizers import PlotNd
from examples import PWSImagePath
import matplotlib.pyplot as plt

plt.ion()  # Without this we will get a crash when trying to open the PlotNd window because a Qt application loop must be running.
plt.figure()

acquisition = pwsdt.Acquisition(PWSImagePath)  # Get a reference to the top-level folder for the measurement.

roiSpecs = acquisition.getRois()  # Get a list of the (name, number, file format) of the available ROIs.
print("ROIs:\n", roiSpecs)

analysis = acquisition.pws.loadAnalysis(acquisition.pws.getAnalyses()[0])  # Load a reference to an analysis file.
kCube = analysis.reflectance  # Load the processed `reflectance` array from the analysis file.

opd, opdValues = kCube.getOpd(useHannWindow=False, indexOpdStop=50)  # Use FFT to transform the reflectance array to OPD

# Scale the opdValues to give estimated depth instead of raw OPD. Factor of 2 because light is making a round-trip.
ri = 1.37  # Estimated RI of livecell chromatin
opdValues = opdValues / (2 * ri)

plotWindow = PlotNd(opd, names=('y', 'x', 'depth'),
                    indices=(None, None, opdValues), title="Estimated Depth")



# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.



import pickle
import numpy as np
from skimage import morphology
import matplotlib.pyplot as plt
from skimage import filters
import multiprocessing as mp
import pwspy.dataTypes as pwsdt
from examples.findOPDSurface.activeContour.funcs import volume3Dto2D, termSeg, morphSmoothing3D, equalAxis3dPlot
import scipy as sp
from glob import glob
import os
from matplotlib import cm


def seg(cellDir: str):
    """This function is run on multiple cores at once so that the full procedure finishes faster.

    Args:
        cellDir: The file path to a PWS acquisition.

    Returns:
        A tuple containing:
            snake: The result of the segmentation. 3d boolean array.
            snake2: The result of the segmentation with morphological smoothing applied. 3d booleaan array.
            opdIndex: The 1d array giving the OPD value at each slice along the 3rd axis of `snake`.

    """
    try:
        anName = 'p0'  # The name of the analysis to load.
        opdCutoffLow = 0  # In preprocessing data below this OPD value will be excluded
        opdCutoffHigh = 25  # In preprocessing data above this OPD value will be excluded

        acq = pwsdt.Acquisition(cellDir) # Load a handle to the acquisition
        an = acq.pws.loadAnalysis(anName)  # Load the analysis results object
        opd, opdIndex = an.reflectance.getOpd(useHannWindow=True)  # Using the Hann window should improve dynamic range and reduce false peaks from frequency leakage.

        # Remove the low and high OPD signal.
        idxLow = np.argmin(np.abs(opdIndex - opdCutoffLow))
        idxHigh = np.argmin(np.abs(opdIndex - opdCutoffHigh))
        opd = opd[:, :, idxLow:idxHigh]
        opdIndex = opdIndex[idxLow:idxHigh]

        # Use RMS ot generate a mask showing where the cells are.
        signal = an.rms  # RMS. Use it to detect background
        thresh = filters.threshold_li(signal)  # Threshold between high-signal and low-signal regions.
        mask = signal < thresh  # Mask that should indicate all regions outside a cell.
        disk = morphology.disk(3)
        mask = morphology.binary_opening(mask, disk)  # Get rid of small isolated regions in the mask.

        #Use the mask from the previous step to generate an rough initial guess for the segmentation. This will be our starting point to save us some iterations.
        level_set = (~mask[:, :, None]) * np.ones(opd.shape)
        level_set[:, :, opd.shape[2] // 2:] = False

        # Blur laterally to denoise. This step is vital (I think)
        sigma = 0.5
        Sigma = sigma / acq.pws.pixelSizeUm  # convert from microns to pixels
        for i in range(opd.shape[2]):
            opd[:, :, i] = sp.ndimage.filters.gaussian_filter(opd[:, :, i], Sigma, mode='reflect')

        # Try to account for decreased sensitivity at higher opd. Amplify higher OPD signal. This is also vital and it's not clear what the exponent should be, they all give different results.
        print("Adjusting OPD signal")
        opd = opd * (opdIndex[None, None, :] ** 0.5)

        opdSum = opd[:, :, ::-1].cumsum(axis=2)[:, :, ::-1]  # Cumulative sum of spectra from high to low opd. This way all pixels inside the cell should be higher values than outside the cell.
        bgSpec = opdSum[mask]  # The signals of the background regions.
        bgSpec = bgSpec.mean(axis=0)  # The average background signal
        opdSum = opdSum - bgSpec[None, None, :]  # Subtract the average background noise from all spectra.

        del opd, bgSpec
        snake = termSeg(opdSum, smoothing=1, lambda2=2, init_level_set=level_set)
        del opdSum
        print("smoothing")
        snake2 = morphSmoothing3D(snake, 5)

        return snake, snake2, opdIndex
    except Exception as e:
        return e


if __name__ == '__main__':
    read = True  # If this is true then we will load result from a previous run and plot them. If False then just run the code and save the results.
    pickleName = r'C:\Users\backman05\Desktop\findsurface\ac9.pickle'  # This is the filename for the results to be saved/loaded from

    if read:
        with open(pickleName, 'rb') as f:
            out = pickle.load(f)

        for i in out:
            if isinstance(i, Exception):
                print(i)
                continue
            snake, smoothed, opdIndex = i

            # Convert from OPD to microns
            height = volume3Dto2D(snake, opdIndex)  # Condense the 3d bool array to a 2d height map
            height = height / 2 / 1.37

            fig, ax = plt.subplots()
            im = plt.imshow(height, cmap='nipy_spectral', clim=[0, 7])
            ax.set_axis_off()
            cbar = plt.colorbar()
            cbar.ax.set_ylabel('microns', rotation=270)
            fig.show()



            # height = volume3Dto2D(smoothed, opdIndex)
            # plt.figure()
            # plt.imshow(height)

            # dx = 0.13 # This is the pixel size. We use this to plot XYZ with equal scaling
            #
            # x, y = np.arange(snake.shape[1]) * dx, np.arange(snake.shape[0]) * dx
            #
            # fig = plt.figure()  # 3D plot
            # ax = fig.add_subplot(projection='3d')
            # equalAxis3dPlot(x, y, height, ax)
            # contour = ax.contourf(x, y, height, levels=50, cmap=cm.nipy_spectral)
            # fig.show()
            # plt.colorbar(contour)

            # #Better looking, poor performance
            # fig = plt.figure()
            # ax = fig.add_subplot(projection='3d')
            # equalAxis3dPlot(x, y, height, ax)
            # X, Y = np.meshgrid(x, y)
            # ax.plot_surface(X, Y, height, cstride=5, rstride=5, cmap=cm.coolwarm)
            # fig.show()
        plt.show()
    else:
        files = glob(os.path.join(r'G:\Data\NA_i_vs_NA_c\matchedNAi_largeNAc\cells', "Cell*"))  # Find all acquisitions in the folder
        with mp.Pool(4) as p:
            out = p.starmap(seg, [(f,) for f in files])  # Run seg for each file, run in parallel for much faster completion.

        with open(pickleName, 'wb') as f:  # Save the results.
            pickle.dump(out, f)





# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

# -*- coding: utf-8 -*-

"""

.. autosummary::
   :toctree: generated/

   analysis
   dataTypes
   utility

"""
import logging
try:
    from . import version
    _versionStr = version.version
except ImportError:  # When running from source the version.py file won't have been created by setuptools_scm
    logging.getLogger(__name__).info("Failed to import `version.py`. Trying to import setuptools_scm")
    try:
        import setuptools_scm
        _versionStr = setuptools_scm.get_version(root='../..', relative_to=__file__)
    except ImportError:  # setuptools_scm
        logging.getLogger(__name__).warn("Failed to import setuptools_scm. Using fallback version string.")
        _versionStr = "x.x.x-dev"

__author__ = 'Nick Anthony'
__version__ = _versionStr

dateTimeFormat = "%d-%m-%Y %H:%M:%S"


__all__ = ['dateTimeFormat', 'utility', 'dataTypes', 'analysis']





# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

# -*- coding: utf-8 -*-
"""
This is a collection of example usages of the :mod:`pwspy` module.

.. autosummary::
    :toctree: generated/

    flatFieldVisualizer
    limitedOPDSigma
    plotITOThinFilmCalibration
    pwsAnalysis
    ROItoMirror
    syntheticMirror
    findOPDSurface
    roiUsageExample
    opdExampleScript
    positionTransformation
"""
import pathlib as pl

# Set this to a folder containing multiple "Cell{x}" acquisition folders. Some examples will use this.
PWSExperimentPath = pl.Path(__file__).parent.parent / 'tests' / 'resources' / 'test_data' / 'sequencer'

# Set this to the "Cell{X}" folder of a PWS acquisition. Most of the examples will then refer to this file.
PWSImagePath = PWSExperimentPath / 'Cell1'



from pwspy.utility.acquisition import loadDirectory, PositionsStep
from pwspy.utility.micromanager import PositionList


class TestSequence:
    def test_sequence(self, sequenceData):
        """Test that the metadata files saved by the event sequencer plugin can be loaded. Use the sequence metadata to load acquisitions
        and a position list giving the locations of each acquisition. Test basic operations on the position list."""
        seq, acqs = loadDirectory(sequenceData.datasetPath)

        seq.printSubTree()

        multiplePosStep: PositionsStep = [i for i in seq.iterateChildren() if isinstance(i, PositionsStep)][0]
        positionDict = multiplePosStep.settings['posList']
        posList = PositionList.fromDict(positionDict)

        assert len(posList) == len(acqs)

        posList2 = posList.copy()
        posList2.mirrorX().mirrorY()  # inplace
        afTransform = posList.getAffineTransform(posList2)
        posList3 = posList.applyAffineTransform(afTransform)

        for acq in acqs:
            iterationNum = acq.sequencerCoordinate.getStepIteration(multiplePosStep)
            print(posList[iterationNum])



from pwspy import analysis
import pwspy.dataTypes as pwsdt
from pwspy.utility.reflection import Material
import pytest
from conftest import testDataPath
import numpy as np

_analysisName = 'testAnalysis'

erMeta = pwsdt.ERMetaData.fromHdfFile(testDataPath / 'extraReflection', 'LCPWS2_100xpfs-8_4_2021')


class TestAnalysis:
    """
    Test the code under pwspy.analysis
    """

    @pytest.mark.parametrize('extraReflection', [None, erMeta])
    def test_pws_analysis(self, dynamicsData, extraReflection):
        """Test that PWS data can be analyzed. Results can be loaded. TODO check that values of analysis results don't change."""
        settings = analysis.pws.PWSAnalysisSettings.loadDefaultSettings("Recommended")

        refAcq = pwsdt.Acquisition(dynamicsData.referenceCellPath)
        ref = refAcq.pws.toDataClass()
        anls = analysis.pws.PWSAnalysis(settings=settings, extraReflectance=extraReflection, ref=ref)

        acq = pwsdt.Acquisition(dynamicsData.datasetPath / "Cell1")
        cube = acq.pws.toDataClass()
        results, warnings = anls.run(cube)

        acq.pws.saveAnalysis(results, _analysisName, overwrite=True)
        with pytest.raises(OSError):
            acq.pws.saveAnalysis(results, _analysisName)  # The analysis already exists so an OSError should be thrown.

        result = acq.pws.loadAnalysis(_analysisName)

        assert isinstance(result.rms, np.ndarray)
        assert isinstance(result.meanReflectance, np.ndarray)
        assert isinstance(result.reflectance, pwsdt.KCube)

    @pytest.mark.parametrize('extraReflection', [None, erMeta])
    def test_dynamics_analysis(self, dynamicsData, extraReflection):
        """Test that dynamics data can be analyzed, results can be loaded"""
        settings = analysis.dynamics.DynamicsAnalysisSettings(
            cameraCorrection=None,
            extraReflectanceId=extraReflection.idTag if extraReflection is not None else None,
            referenceMaterial=Material.Water,
            numericalAperture=0.52,
            relativeUnits=True
        )

        refAcq = pwsdt.Acquisition(dynamicsData.referenceCellPath)
        ref = refAcq.dynamics.toDataClass()
        anls = analysis.dynamics.DynamicsAnalysis(settings=settings, extraReflectance=extraReflection, ref=ref)

        acq = pwsdt.Acquisition(dynamicsData.datasetPath / 'Cell1')
        cube = acq.dynamics.toDataClass()
        results, warnings = anls.run(cube)

        acq.dynamics.saveAnalysis(results, _analysisName, overwrite=True)
        with pytest.raises(OSError):
            acq.dynamics.saveAnalysis(results, _analysisName)  # The analysis already exists so an OSError should be thrown.

        result = acq.dynamics.loadAnalysis(_analysisName)

        assert isinstance(result.rms_t_squared, np.ndarray)
        assert isinstance(result.meanReflectance, np.ndarray)
        assert isinstance(result.reflectance, pwsdt.DynCube)

    def test_compilation(self, dynamicsData):
        """Test that ROIs and PWS/Dynamics analysis results can be successfully `compiled` into a data table."""
        settings = analysis.compilation.GenericCompilerSettings(roiArea=True)
        genComp = analysis.compilation.GenericRoiCompiler(settings)

        settings = analysis.compilation.PWSCompilerSettings(reflectance=True, rms=True, polynomialRms=True)
        pwsComp = analysis.compilation.PWSRoiCompiler(settings)

        settings = analysis.compilation.DynamicsCompilerSettings(meanReflectance=True, rms_t_squared=True, diffusion=True)
        dynComp = analysis.compilation.DynamicsRoiCompiler(settings)

        acq = pwsdt.Acquisition(dynamicsData.datasetPath / 'Cell1')

        results = []
        for roiSpecs in acq.getRois():
            roi = acq.loadRoi(*roiSpecs)
            results.append((genComp.run(roi),
                       pwsComp.run(acq.pws.loadAnalysis(_analysisName), roi.getRoi()),
                       dynComp.run(acq.dynamics.loadAnalysis(_analysisName), roi.getRoi())))

            print(f"Successfully Compiled {len(results)} ROIs for general, PWS, and dynamics analysis.")





# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

# -*- coding: utf-8 -*-
"""
This file is used to install the pwspy package. for example navigate in your terminal to the directory containing this
file and type `pip install .`. This file is also used by the Conda recipe (buildscripts/conda)
"""
from setuptools import setup, find_packages
import setuptools_scm

from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(name='pwspy',
      version=setuptools_scm.get_version(write_to="src/pwspy/version.py"),
      description='A framework for working with Partial Wave Spectroscopic microscopy files.',
      long_description=long_description,
      long_description_content_type='text/markdown',
      author='Nick Anthony',
      author_email='nicholas.anthony@northwestern.edu',
      url='https://github.com/BackmanLab/PWSpy',
      python_requires='>=3.7',
      install_requires=['numpy',
                        'scipy',
                        'matplotlib',
                        'tifffile',
                        'psutil',
                        'shapely',
                        'pandas',
                        'h5py',
                        'jsonschema',
                        'opencv-python', #opencv is required but naming differences between conda and pip seem to cause issues. Maybe should be commented out?
                        'scikit-image',
                        'rasterio'],
      package_dir={'': 'src'},
      package_data={'pwspy': ['utility/reflection/refractiveIndexFiles/*',
                              'utility/thinFilmInterferenceFiles/*',
                              'analysis/_resources/defaultAnalysisSettings/*',
                              'dataTypes/jsonSchemas/*']},
      packages=find_packages('src')
	)



# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

from pwspy.dataTypes import DynCube
import os
from glob import glob

wDir = r''
refPath = r''

ref = DynCube.loadAny(refPath)
ref.correctCameraEffects()
ref.normalizeByExposure()

files = glob(os.path.join(wDir, 'Cell*'))
for f in files:
    dyn = DynCube.loadAny(f)
    dyn.correctCameraEffects()
    dyn.normalizeByExposure()
    dyn.normalizeByReference(ref)

    # The original MATLAB script optionally uses 3 frame frame-averaging here as a lowpass, That hasn't been implemented here

    #This is equivalent to subtracting the mean from each spectra and taking the RMS
    rms = dyn.data.std(axis=2)

    #TODO save the RMS


from __future__ import annotations
import typing


class TreeItem:
    """Basic implementation of an item for a tree. Our `treemodel` subclassed from QAbstractItemModel is designed to work with this class and it's subclasses."""
    def __init__(self):
        self._parentItem = None
        self._itemData = {}
        self._childItems: typing.List[TreeItem] = []

    def addChild(self, item: TreeItem):
        item._parentItem = self
        self._childItems.append(item)

    def addChildren(self, children: typing.Sequence[TreeItem]):
        for i in children:
            i._parentItem = self
        self._childItems.extend(children)

    def child(self, row) -> TreeItem:
        return self._childItems[row]

    def children(self) -> typing.Tuple[TreeItem]:
        return tuple(self._childItems)

    def childCount(self) -> int:
        return len(self._childItems)

    def parent(self) -> TreeItem:
        return self._parentItem

    def row(self) -> int:
        """Return which row we are with respect to the parent."""
        if self._parentItem:
            return self._parentItem._childItems.index(self)
        else:
            return None

    def iterateChildren(self) -> typing.Generator[TreeItem]:
        """Recursively iterate through all children of this step"""
        for child in self._childItems:
            yield child
            yield from child.iterateChildren()

    def getTreePath(self) -> typing.Tuple[TreeItem]:
        """Return a list of steps starting with the root step and ending with this step."""
        l: typing.List[TreeItem] = []
        step = self
        while True:
            l.append(step)
            if step._parentItem is None:
                break
            step = step._parentItem
        return tuple(reversed(l))

    def data(self, role: int) -> typing.Any:
        try:
            return self._itemData[role]
        except KeyError:
            return None

    def setData(self, role: int, data: typing.Any):
        self._itemData[role] = data




# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

# -*- coding: utf-8 -*-
"""
Custom datatypes that are commonly used in the analysis of PWS related data.


Metadata Classes
------------------
These classes provide handling of information about an acquisition without requiring that the full data be loaded into RAM. These can be used to get information
about the equipment used, the date the acquisition was taken, the location of the files, the presence of ROIs or analyses, etc.

.. autosummary::
    :toctree: generated/
    :nosignatures:

    PwsMetaData
    DynMetaData
    ERMetaData
    FluorMetaData

Data Classes
-----------------
These classes are used to actuallly load and manipulate acquisition data. The all have a corresponding metadata class.

.. autosummary::
    :toctree: generated/
    :nosignatures:

    PwsCube
    DynCube
    KCube
    ExtraReflectanceCube
    ExtraReflectionCube
    ICBase
    ICRawBase

Other Classes
---------------
.. autosummary::
    :toctree: generated/
    :nosignatures:

    Roi
    RoiFile
    CameraCorrection
    Acquisition
    FluorescenceImage

Inheritance
-------------
.. inheritance-diagram:: PwsCube DynCube PwsMetaData DynMetaData ExtraReflectionCube ExtraReflectanceCube KCube FluorMetaData
    :top-classes: ICBase, MetaDataBase, AnalysisManager, ICRawBase
    :parts: 1
    :caption: Abstract base classes define common behavior between the implementations of the various data types making it easy to write software using PWSpy that will work for all available data type implementations.

"""

import os
_jsonSchemasPath = os.path.join(os.path.split(__file__)[0], 'jsonSchemas')
from ._metadata import (PwsMetaData, Acquisition, DynMetaData, ERMetaData, FluorMetaData, AnalysisManager, MetaDataBase,
                        MetaDataBase)
from ._other import Roi, CameraCorrection, RoiFile
from ._data import (FluorescenceImage, ExtraReflectanceCube, ExtraReflectionCube, PwsCube, KCube, DynCube, ICBase,
                    ICRawBase)

__all__ = ['PwsMetaData', 'Acquisition', 'DynMetaData', 'ERMetaData', 'FluorMetaData', 'AnalysisManager', 'MetaDataBase',
           'MetaDataBase', 'Roi', 'CameraCorrection', 'FluorescenceImage', 'ExtraReflectionCube',
           'ExtraReflectanceCube', 'PwsCube', 'KCube', 'DynCube', 'ICBase', 'ICRawBase', 'RoiFile']









# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

"""Backing functions used by the activeContour script."""
import typing
import numpy as np
from skimage import morphology, segmentation
import matplotlib.pyplot as plt
import scipy.ndimage as ndi

# Convert snake to height map
def volume3Dto2D(arr3d, valueIndex=None, fillValue=0):
    """Convert a 3d boolean volume to a 2d array with values giving the height of the volume"""
    if valueIndex is None:
        valueIndex = np.arange(arr3d.shape[2])
    height = valueIndex[::-1][np.argmax(arr3d[:, :, ::-1], axis=2)]  # Get opdIndex value corresponding to highest True value in 3d array.
    height[~np.any(arr3d, axis=2)] = fillValue  # The above step doesn't work if no True values are found. Set pixels that are completely False to 0.
    return height


def equalAxis3dPlot(X: typing.Sequence[float], Y: typing.Sequence[float], Z: typing.Sequence[float], ax: plt.Axes):
    """With no built-in way to set the aspect of a 3d plot we force it with this. This function finds the bounding box
    of X, Y, and Z and then sets the limits of `ax` accordingly.

    Args:
        X: Sequence of x coordinates of data
        Y: Sequence of y coordinates of data
        Z: Sequence of z coordinates of data
        ax: The `Axes` that will be plotted to.

    """
    # Create cubic bounding box to simulate equal aspect ratio
    max_range = np.array([X.max() - X.min(), Y.max() - Y.min(), Z.max() - Z.min()]).max()
    Xb = 0.5 * max_range * np.mgrid[-1:2:2, -1:2:2, -1:2:2][0].flatten() + 0.5 * (X.max() + X.min())
    Yb = 0.5 * max_range * np.mgrid[-1:2:2, -1:2:2, -1:2:2][1].flatten() + 0.5 * (Y.max() + Y.min())
    Zb = 0.5 * max_range * np.mgrid[-1:2:2, -1:2:2, -1:2:2][2].flatten() + 0.5 * (Z.max() + Z.min())
    # Comment or uncomment following both lines to test the fake bounding box:
    for xb, yb, zb in zip(Xb, Yb, Zb):
        ax.plot([xb], [yb], [zb], 'w')


def morphSmoothing3D(arr: np.ndarray, radius: int):
    """Smooth a 3d boolean array along all 3 axes. Warning this is way too slow for radius > 10."""
    b = morphology.ball(radius)  # A 3d sphere structuring element.
    # #So Slow
    # arr = morphology.closing(arr, b)
    # arr = morphology.opening(arr, b)
    arr = ndi.binary_closing(arr, b)
    arr = ndi.binary_opening(arr, b)
    return arr


def morphSmoothing2D(arr: np.ndarray, radius: int):
    """Smooth a 3d boolean array only along the first and second axes. Warning this is way too slow for radius > 10."""
    b = morphology.disk(radius)  # A 3d sphere structuring element.
    arr = morphology.binary_closing(arr, b[:, :, None])
    arr = morphology.binary_opening(arr, b[:, :, None])
    return arr


def terminableSnake(arr: np.ndarray, smoothing=1, lambda1=1, lambda2=1, init_level_set='checkerboard'):
    """The morphological_chan_vese algorithm in scikit-image runs for a fixed number of iterations with no detection of when the process should terminate.
    This function attempts to detect when the algorithm is done and then forces it to terminate. Check the scikit-image documentation for a description
    of the input arguments and return values."""

    class EndItException(Exception):
        """We throw this exception to prematurely exit out of scikit-image's morphological_chan_vese algorithm."""
        pass

    def saveEvolution() -> typing.Tuple[typing.Callable, list, list]:
        """Generates a callback function  which is passed as `iter_callback` of the morphological_chan_vese algorithm.

        Returns: A tuple containing:
            A function to be passed as `iter_callback` to the morpholocial_chan_vese algorithm. It will throw an `EnditException`
            end the algorithm when it has determined that the algorithm has converged on a solution.

            A list containing the result of the algorithm after each iteration. Used to investigate the progress of the algorithm.

            A list containing the number of pixels that changed from false to true during each iteration. This is used to determine when a solution has started
            converging.
        """
        lst = []
        changes = []
        iteration = [0] # Just used to store an integer of how many times the function has been run.

        def _store(x: np.ndarray): # This function gets run at the end of each algorithm iteration. `x` is the current segmentation result.
            print(f'iter {iteration[0]}')  # Print out the iterationso we can view how fast we are proceeding.
            iteration[0] += 1
            if iteration[0] > 1:
                change = (x - lst[-1]).sum() / x.size  # calculate the ratio of pixels that have changed. note: this is signed, so if 10 pixel go from True to False and 10 pixels go from False to True we end up with 0.
                if len(changes) > 0:
                    if change * changes[-1] < 0: #If the sign of change has switched then we've gone from the segmentation growing to shrinking or vice-versa. The algorithm will alternate back and forth here, We use this as an indicator that the algorithm can be terminated.
                        lst.append(np.copy(x))  # Make sure to save our most recent result before exiting.
                        raise EndItException("Sign change")
                changes.append(change) # Store the change from this iteration to the last.
            lst.append(np.copy(x))

        return _store, lst, changes

    try:
        cb, l, changes = saveEvolution()
        snake = segmentation.morphological_chan_vese(arr, 100, smoothing=smoothing, lambda1=lambda1, lambda2=lambda2, iter_callback=cb,
                                                     init_level_set=init_level_set)
    except EndItException as e:
        print(e)
        snake = l[-1]
    return snake


def termSeg(arr: np.ndarray, smoothing=1, lambda1=1, lambda2=1, init_level_set='checkerboard'):
    """Start with a downscaled image and then move back up to full resoltuion to much more quickly segment an image.
    Check the scikit-image documentation for a description of the input arguments and return values."""
    from skimage.transform import downscale_local_mean, rescale

    if isinstance(init_level_set, np.ndarray):
        init_level_set = downscale_local_mean(init_level_set, (8, 8, 1)).astype(bool)

    img = downscale_local_mean(arr, (8, 8, 1))  # downscale by 8x in the x/y plane.
    snake = terminableSnake(img, smoothing, lambda1, lambda2, init_level_set)  # Segment the downscaled image.

    img = downscale_local_mean(arr, (2, 2, 1))  # downscale the original image by 2x in the x/y plane.
    snake = rescale(snake, (4, 4, 1), preserve_range=True, order=0).astype(bool)  # Upscale the previous segmentation result and use it as our starting point
    snake = terminableSnake(img, smoothing, lambda1, lambda2, snake) # Segment at 2x downscale.

    snake = rescale(snake, (2, 2, 1), preserve_range=True, order=0).astype(bool)  # Upscale the previous result to full resolution.
    snake = terminableSnake(arr, smoothing, lambda1, lambda2, snake)  # Use the previous result as a starting point for a final sementation at full resolution.

    return snake



# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

from pwspy.dataTypes import DynCube, Roi
import os
import numpy as np
import pwspy.dataTypes as pwsdt

wDir = r''
cellNum = list(range(1001, 1015))  # Cell numbers to analyze
mirrorNum = 937  # Flat normalization cube, this has been processed by `timeseries2imagecube_general`
background = 1997  # Normal cube used for temporal background for noise subtraction

# It's not clear why there is a mirror and a background. Couldn't you just have one. You could then use the mean of the
# data as the mirror while still extracting noise data.

roiName = 'nuc'

refAcq = pwsdt.Acquisition(os.path.join(wDir, f'DYN_Cell{mirrorNum}'))
ref = refAcq.dynamics.toDataClass()

for num in cellNum + [background]: #Loop through all cell folders +1 for background
    dynAcq = pwsdt.Acquisition(os.path.join(wDir, f'DYN_Cell{num}'))
    dyn = dynAcq.dynamics.toDataClass()
    dyn.correctCameraEffects()
    dyn.normalizeByExposure()
    dyn.normalizeByReference(ref)
    dyn.data = dyn.data - dyn.data.mean(axis=2)[:, :, None] #Subtract the mean of each spectra

    if num == background: #If background then generate a random roiFile
        # Randomly sample points of background to save space
        sampleSize = 30000
        rows = np.random.choice(dyn.data.shape[0], sampleSize)
        cols = np.random.choice(dyn.data.shape[1], sampleSize)
        mask = np.zeros(dyn.data.shape).astype(np.bool)
        mask[zip(rows, cols)] = True
        rois = [Roi.fromMask(mask)]
    else:
        rois = [dynAcq.loadRoi(roiName, roiNum) for name, roiNum, _ in dynAcq.getRois() if name == roiName]

    for roi in rois:
        F = np.fft.rfft(dyn.data[roi.mask], axis=1)  # FFT of each spectra in the roiFile
        data = np.fft.irfft(F*np.conjugate(F), axis=1) / F.shape[1]  # Autocorrelation
        truncLength = 100
        data = data[:, :truncLength]
        # TODO save the result